from .types import AdditionalType, OptionType
from ..tools import get, setup_logger
from ..errors import CouldNotParse

import discord
from discord.enums import ChannelType

import typing

logging = setup_logger(__name__)

class ParseMethod:
    """Methods of how the interaction argument data should be treated

    - ``RAW``       [0]
        Returns the raw value which was received
    - ``RESOLVE``   [1]
        Uses the resolved data which will be delivered together with the received interaction
    - ``FETCH``     [2]
        Fetches all the ids of the received data with an api call
    - ``CACHE``     [3]
        Uses the internal bot cache to get the data 

    .. warning::

        The cache method uses the `.get_guild`, `.get_channel`... methods, which needs to have some intents enabled 
        (`more information <https://discordpy.readthedocs.io/en/latest/intents.html?highlight=intents>`__)

    - ``AUTO``      [4]
        This will try all methods beginning (RESOLVE, FETCH, CACHE, RAW) and changes to the next method whenever an exception occurs
    """
    
    RAW         =       Raw         =       0
    RESOLVE     =       Resolve     =       1
    FETCH       =       Fetch       =       2
    CACHE       =       Cache       =       3
    AUTO        =       Auto        =       4

def format_name(value):
    return str(value).lower().replace(" ", "-")

def resolve(data, _state):
    resolved = {}
    for x in data["data"]["resolved"]:
        if x == "members":
            resolved["members"] = {}
            for m_id in data["data"]["resolved"]["members"]:
                member_data = data["data"]["resolved"]["members"][m_id]
                member_data["user"] = data["data"]["resolved"]["users"][m_id]
                resolved["members"][m_id] = discord.Member(data=member_data, guild=_state._get_guild(int(data["guild_id"])), state=_state)
        elif x == "messages":
            resolved["messages"] = {}
            for message_id in data["data"]["resolved"]["messages"]:
                message_data = data["data"]["resolved"]["messages"][message_id]
                resolved["messages"][message_id] = discord.Message(data=message_data, channel=_state.get_channel(data["channel_id"]), state=_state)
        elif x == "channels":
            resolved["channels"] = {}
            for channel_id in data["data"]["resolved"]["channels"]:
                channel_data = data["data"]["resolved"]["channels"][channel_id]

                guild = _state._get_guild(data["guild_id"])
                channel = None
                if ChannelType(channel_data["type"]) is ChannelType.text:
                    channel = discord.TextChannel(data=channel_data, guild=guild, state=_state)
                elif ChannelType(channel_data["type"]) is ChannelType.voice:
                    channel = discord.VoiceChannel(data=channel_data, guild=guild, state=_state)
                elif ChannelType(channel_data["type"]) is ChannelType.category:
                    channel = discord.CategoryChannel(data=channel_data, guild=guild, state=_state)
                elif ChannelType(channel_data["type"]) is ChannelType.group:
                    channel = discord.GroupChannel(data=channel_data, guild=guild, state=_state)
                elif ChannelType(channel_data["type"]) is ChannelType.news:
                    channel = discord.NewsChannel(data=channel_data, guild=guild, state=_state)
                elif ChannelType(channel_data["type"]) is ChannelType.private:
                    channel = discord.DMChannel(data=channel_data, guild=guild, state=_state)
                elif ChannelType(channel_data["type"]) is ChannelType.store:
                    channel = discord.StoreChannel(data=channel_data, guild=guild, state=_state)
                elif ChannelType(channel_data["type"]) is ChannelType.stage_voice:
                    channel = discord.StageChannel(data=channel_data, guild=guild, state=_state)
                resolved["channels"][channel_id] = channel
        elif x == "roles":
            resolved["roles"] = {}
            for role_id in data["data"]["resolved"]["roles"]:
                role_data = data["data"]["resolved"]["roles"][role_id]
                resolved["roles"][role_id] = discord.Role(data=role_data, guild=_state._get_guild(data["guild_id"]), state=_state)
        elif x == "users":
            pass
        else:
            logging.warning("Could not resolve data of type '" + str(x) +  "'")

    return resolved

async def fetch_data(value, typ, data, _discord):
    logging.debug("fetching something with type " + str(typ) + " value " + str(value))
    if typ == OptionType.MEMBER:
        return await (await _discord.fetch_guild(int(data["guild_id"]))).fetch_member(int(value))
    elif typ == OptionType.CHANNEL:
        return await _discord.fetch_channel(int(value))
    elif typ == OptionType.ROLE:
        return get(await (await _discord.fetch_guild(int(data["guild_id"]))).fetch_roles(), value, lambda x: getattr(x, "id", None) == int(value))
    elif typ == AdditionalType.MESSAGE:
        return await (await _discord.fetch_channel(data["channel_id"])).fetch_message(int(value))
    else:
        return value

def resolve_data(value, typ, data, state):
    resolved = resolve(data, state)
    logging.debug("resolving something with type " + str(typ) + " value " + str(value))
    if typ == OptionType.MEMBER:
        return resolved["members"].get(value)
    elif typ == OptionType.ROLE:
        return resolved["roles"].get(value)
    elif typ == OptionType.CHANNEL:
        return resolved["channels"].get(value)
    elif typ == AdditionalType.MESSAGE:
        return resolved["messages"].get(value)
    else:
        return value

def cache_data(value, typ, data, _state):
    logging.debug("getting something out of the cache with type " + str(typ) + " value " + str(value))
    if typ in [OptionType.STRING, OptionType.INTEGER, OptionType.BOOLEAN, OptionType.FLOAT]:
        return value
    elif typ == OptionType.MEMBER:
        return _state._get_guild(int(data["guild_id"])).get_member(int(value))
    elif typ == OptionType.CHANNEL:
        return _state.get_channel(int(value))
    elif typ == OptionType.ROLE:
        return _state._get_guild(int(data["guild_id"])).get_role(int(value))
    elif typ == AdditionalType.MESSAGE:
        return _state._get_guild(int(data["guild_id"])).get_partial_message(int(value))
    elif typ == AdditionalType.GUILD:
        return _state._get_guild(int(value))
    else:
        return value

async def handle_options(data, options, method, _discord: discord.Client):
    _options = {}
    for op in options:
        if op["type"] not in [OptionType.SUB_COMMAND, OptionType.SUB_COMMAND_GROUP]:
            parsed = await handle_thing(op["value"], op["type"], data, method, _discord)
            logging.debug("value in handle_options is " + str(op["value"]) + " with type " + str(op["type"]) + " and name is " + str(op["name"]) + " parsed " + str(parsed))
            
            if parsed is None:
                raise CouldNotParse(op["value"], op["type"], method)
            _options[op["name"]] = parsed
    return _options

async def handle_thing(value, typ, data, method, _discord, auto=False) -> typing.Union[str, int, bool, discord.Member, discord.TextChannel, discord.Role, float, discord.Message, discord.Guild]:
    logging.debug("Trying to handle val " + str(value) + " type " + str(typ) +  " with method " + str(method) + " auto is" + str(auto))
    if method is ParseMethod.RESOLVE or method is ParseMethod.AUTO:
        try:
            return resolve_data(value, typ, data, _discord._connection)
        except Exception as ex:
            logging.warning("Got exepction while resolving data" +
                f"\n{type(ex).__name__}: {ex}\n" +
                f"{__file__}:{ex.__traceback__.tb_lineno}" +
                ("\nTrying next method" if method is ParseMethod.AUTO else "")
            )
            if method is ParseMethod.AUTO:
                return await handle_thing(value, typ, data, ParseMethod.FETCH, _discord, True)
    elif method is ParseMethod.FETCH:
        try:
            return await fetch_data(value, typ, data, _discord)
        except Exception as ex:
            logging.warning("Got exepction while getting data from cache" +
                f"\n{type(ex).__name__}: {ex}\n" +
                f"{__file__}:{ex.__traceback__.tb_lineno}" +
                ("\nTrying next method" if method is ParseMethod.AUTO else "")
            )
            if auto is True:
                return await handle_thing(value, typ, data, ParseMethod.CACHE, _discord, auto)
    elif method is ParseMethod.CACHE:
        try:
            return cache_data(value, typ, data, _discord._connection)
        except Exception as ex:
            logging.warning("Got exepction while resolving data" +
                f"\n{type(ex).__name__}: {ex}\n" +
                f"{__file__}:{ex.__traceback__.tb_lineno}" +
                ("\nTrying next method" if method is ParseMethod.AUTO else "")
            )
            if auto:
                return await handle_thing(value, typ, data, ParseMethod.RAW, _discord, auto)
    elif method is ParseMethod.RAW:
        return value
    else:
        logging.warning("Unkonw parsemethod: " + str(method) + "\nReturning raw value")
        return value

def create_choice(name, value) -> dict:
    """A function that will create a choice for a :class:`~SlashOption`
    
    Parameters
    ----------
        name: :class:`str`
            The name of the choice
        value: :class:`Any`
            The value that will be received when the user selected this choice
    
    Returns
    -------
        :returns: The created choice
        :type: :class:`dict`
    
    """
    return {"name": name, "value": value}


from .tools import MISSING
from .errors import InvalidLength, OutOfValidRange, WrongType

from discord import Emoji
from discord.errors import InvalidArgument

import inspect
from enum import IntEnum
from typing import Any, List, Union


class SelectOption():
    """
    An option for a select menu
    
    Parameters
    ----------
    value: :class:`str`
        The dev-define value of the option, max 100 characters
    label: :class:`str`
        The user-facing name of the option, max 25 characters; default \u200b ("empty" char)
    description: :class:`str`, optional
        An additional description of the option, max 50 characters
    emoji : :class:`discord.Emoji` | :class:`str`, optional
        Emoji appearing before the label; default MISSING
    """
    def __init__(self, value, label="\u200b", description=None, emoji=None) -> None:
        """
        Creates a new SelectOption

        Example:
        ```py
        SelectOption(label="This is a option", value="my_value", description="This is the description of the option")
        ```
        """
        self._label = None
        self._value = None
        self._description = None
        self._emoji = None
        self.default: bool = False
        """
        Whether this option is selected by default in the menu or not

        :type: :class:`bool`
        """

        self.label = label
        self.value = value
        self.description = description
        self.emoji = emoji
    def __repr__(self) -> str:
        return f"<discord_ui.SelectOption(label={self.label}, value={self.value})>"

    @property
    def content(self) -> str:
        """
        The complete option content, consisting of the emoji and label

        :type: :class:`str`
        """
        return (self.emoji + " ") if self.emoji is not None else "" + (self.label or '')
    
    @property
    def label(self) -> str:
        """
        The main text appearing on the option 

        :type: :class:`str`
        """
        return self._label
    @label.setter
    def label(self, value: str):
        if value is None:
            value = ""
        elif value is not None and not isinstance(value, str):
            raise WrongType("label", value, "str")
        elif value is not None and len(value) > 100 and value > 0:
            raise InvalidLength("label", value, 100, 0)
        self._label = value

    @property
    def value(self) -> Any:
        """
        A unique value for the option, which will be usedd to identify the selected value

        :type: :class:`str`
        """
        return self._value
    @value.setter
    def value(self, value):
        if inspect.isclass(value):
            raise WrongType("value", value, ["int", "str", "bool", "float"])
        if isinstance(value, str):
            if len(value) > 100 or len(value) < 1:
                raise InvalidLength("value", _min=1, _max=100)
        self._value = value

    @property
    def description(self) -> str:
        """
        A description appearing on the option

        :type: :class:`str`
        """
        return self._description
    @description.setter
    def description(self, value):
        if value is not None and not isinstance(value, str):
            raise WrongType("description", "str")
        if value is not None and len(value) > 100:
            raise InvalidLength("description", 100, 0)
        self._description = value
    
    @property
    def emoji(self) -> str:
        """
        The mention of the emoji before the text

            .. note::
                For setting the emoji, you can use a str or discord.Emoji
        
        :type: :class:`str`
        """
        if self._emoji is None:
            return None
        if "id" not in self._emoji:
            return self._emoji["name"]
        return f'<{"a" if "animated" in self._emoji else ""}:{self._emoji["name"]}:{self._emoji["id"]}>'
    @emoji.setter
    def emoji(self, val: Union[Emoji, str, dict]):
        """The emoji appearing before the label"""
        if val is None:
            self._emoji = None
        elif isinstance(val, str):
            self._emoji = {
                "id": None,
                "name": val
            }
        elif isinstance(val, Emoji):
            self._emoji = {
                "id": val.id,
                "name": val.name,
                "animated": val.animated
            }
        elif isinstance(val, dict):
            self._emoji = val
        elif val is None:
            self._emoji = None
        else:
            raise WrongType("emoji", val, ["str", "discord.Emoji", "dict"])


    def to_dict(self) -> dict:
        payload = {
            "label": self._label,
            "value": self._value,
            "default": self.default
        }
        if self._description is not None:
            payload["description"] = self._description
        if self._emoji is not None:
            payload["emoji"] = self._emoji
        return payload

    @classmethod
    def _fromData(cls, data) -> "SelectOption":
        """
        Initializes a new SelectOption from a dict
        
        Parameters
        ----------
            data: :class:`dict`
                The data to initialize from
        Returns
        -------
            :class:`~SelectOption`
                The new Option generated from the dict
        
        """
        x = SelectOption(data["value"], data.get("label"), data.get("description"), data.get("emoji"))
        x.default = data.get("default", False)
        return x

class Component():
    def __init__(self, component_type) -> None:
        self._component_type = getattr(component_type, "value", component_type)
    @property
    def component_type(self) -> int:
        """
        The message component type

        :type: :class:`int`
        """
        return ComponentType(self._component_type)

class UseableComponent(Component):
    def __init__(self, component_type) -> None:
        Component.__init__(self, component_type)
        self._custom_id = None
    @property
    def custom_id(self) -> str:
        """
        The custom_id of the menu to identify it

        :type: :class:`str`
        """
        return self._custom_id
    @custom_id.setter
    def custom_id(self, value: str):
        if len(value) > 100 or len(value) < 1:
            raise InvalidLength("custom_id", 0, 100)
        if not isinstance(value, str):
            raise WrongType("custom_id", value, "str")
        self._custom_id = value

class SelectMenu(UseableComponent):
    """
    Represents a ui-dropdown selectmenu

    Parameters
    ----------
    custom_id: :class:`str`
        The custom_id for identifying the menu, max 100 characters
    options: List[:class:`~SelectOption`]
        A list of options to select from
    min_values: :class:`int`, optional
        The minimum number of items that must be chosen; default ``1``, min 0, max 25
    max_values: :class:`int`, optional
        The maximum number of items that can be chosen; default ``1``, max 25
    placeholder: :class:`str`, optional
        A custom placeholder text if nothing is selected, max 100 characters; default MISSING
    default: :class:`int`, optional
        The position of the option that should be selected by default; default MISSING
    disabled: :class:`bool`, optional
        Whether the select menu should be disabled or not; default ``False``
    """
    def __init__(self, custom_id, options, min_values=1, max_values=1, placeholder=None, default=None, disabled=False) -> None:
        """
        Creates a new ui select menu

        Example:
        ```py
        SelectMenu(custom_id="my_id", options=[SelectOption(...)], min_values=2, placeholder="select something", default=0)
        ```
        """
        UseableComponent.__init__(self, ComponentType.Select)
        self._options = None

        self.max_values: int = 0
        """
        The maximum number of items that can be chosen; default 1, max 25

        :type: :class:`int`
        """
        self.min_values: int = 0
        """
        The minimum number of items that must be chosen; default 1, min 0, max 25

        :type: :class:`int`
        """
        self.disabled = disabled
        """
        Whether the selectmenu is disabled or not

        :type: :class:`bool`
        """
        self.placeholder: str = placeholder
        """
        Custom placeholder text if nothing is selected

        :type: :class:`str` | :class:`None`
        """
        self.custom_id = custom_id
        self.disabled = disabled
        self.options = options

        if min_values is not None and max_values is None:
            if min_values < 0 or min_values > 25:
                raise OutOfValidRange("min_values", 0, 25)
            self.min_values = min_values
            self.max_values = min_values
        elif min_values is None and max_values is not None:
            if max_values < 0 or min_values > 25:
                raise OutOfValidRange("min_values", 0, 25)
            self.min_values = 1
            self.max_values = max_values
        elif min_values is not None and max_values is not None:
            if max_values < 0 or min_values > 25:
                raise OutOfValidRange("min_values", 0, 25)
            if min_values < 0 or min_values > 25:
                raise OutOfValidRange("min_values", 0, 25)
            self.min_values = min_values
            self.max_values = max_values
        
        if default is not None:
            self.set_default_option(default)
    def __str__(self) -> str:
        return self.custom_id
    def __repr__(self) -> str:
        return f"<discord_ui.SelectMenu(custom_id={self.custom_id}, options={self.options})>"
    
    @staticmethod
    def _fromData(data) -> 'SelectMenu':
        return SelectMenu(data["custom_id"], data["options"], data.get("min_values"), data.get("max_values"), data.get("placeholder"), disabled=data.get("disabled", False))
    # region props
    @property
    def options(self) -> List[SelectOption]:
        """
        The options in the select menu to select from

        :type: List[:class:`~SelectOption`]
        """
        return [SelectOption._fromData(x) for x in self._options]
    @options.setter
    def options(self, value: List[SelectOption]):
        if isinstance(value, list):
            if len(value) > 25 or len(value) == 0:
                raise OutOfValidRange("length of options", 1, 25)
            if all(isinstance(x, SelectOption) for x in value):
                self._options = [x.to_dict() for x in value]
            elif all(isinstance(x, dict) for x in value):
                self._options = value
            else:
                raise WrongType("options", value, ["List[SelectOption]", "List[dict]"])
        else:
            raise WrongType("options", value, "list")

    @property
    def default_option(self) -> Union[SelectOption, None]:
        """
        The option selected by default

        :type: :class:`~SelectOption` | :class:`None`
        """
        x = [x for x in self.options if x.default]
        if len(x) == 1:
            return x[0]
    def set_default_option(self, position: int) -> 'SelectMenu':
        """
        Selects the default selected option

        Parameters
        ----------
        position: :class:`int`
            The position of the option that should be default
        """
        if not isinstance(position, int):
            raise WrongType("position", position, "int")
        if position < 0 or position >= len(self.options):
            raise OutOfValidRange("default option position", 0, str(len(self.options) - 1))
        self._options[position]["default"] = True
        return self
    # endregion

    def to_dict(self) -> dict:
        payload = {
            "type": self._component_type,
            "custom_id": self._custom_id,
            "options": self._options,
            "disabled": self.disabled,
            "min_values": self.min_values,
            "max_values": self.max_values
        }
        if self.placeholder is not None:
            payload["placeholder"] = self.placeholder
        return payload

class BaseButton(Component):
    def __init__(self, label, color, emoji, new_line, disabled) -> None:
        Component.__init__(self, ComponentType.Button)
        if label is None and emoji is None:
            raise InvalidArgument("You need to pass a label or an emoji")
        self._label = None
        self._style = None
        self._emoji = None
        self._url = None

        self.new_line = new_line
        self.label = label
        self.color = color
        self.disabled = disabled
        self.emoji = emoji

    def __str__(self) -> str:
        return self.content
    def to_dict(self):
        payload = {"type": self._component_type, "style": self._style, "disabled": self.disabled, "emoji": self._emoji}
        if self._style == ButtonStyles.url:
            payload["url"] = self._url
        else:
            payload["custom_id"] = self._custom_id
        if self._emoji is not None:
            payload["emoji"] = self._emoji
        if self._label is not None:
            payload["label"] = self._label
        return payload

    @property
    def content(self) -> str:
        """
        The complete content in the button ("{emoji} {label}")

        :type: :class:`str`
        """
        return (self.emoji + " " if self.emoji is not None else "") + (self.label or '')
        
    @property
    def label(self) -> str:
        """
        The label displayed on the button

        :type: :class:`str`
        """
        return self._label
    @label.setter
    def label(self, val: str):
        if val is None:
            val = ""
        elif val is not None and not isinstance(val, str):
            raise WrongType("label", val, "str")
        elif val is not None and len(val) > 100:
            raise InvalidLength("label", _max=100)
        elif val is not None and len(val) < 1:
            raise InvalidLength("label", _min=0)

        self._label = str(val)

    @property
    def color(self) -> int:
        """
        The color for the button

        :type: :class:`int`, one of :class:`~ButtonStyles`
        """
        return self._style
    @color.setter
    def color(self, val):
        if ButtonStyles.getColor(val) is None:
            raise InvalidArgument(str(val) + " is not a valid color")
        self._style = ButtonStyles.getColor(val).value
    
    @property
    def emoji(self) -> str:
        """
        The mention of the emoji before the text
        
            .. note::
                For setting the emoji, you can use a str or discord.Emoji          
        
        :type: :class:`str`
        """
        if self._emoji is None:
            return None
        if "id" not in self._emoji:
            return self._emoji["name"]
        return f'<{"a" if "animated" in self._emoji else ""}:{self._emoji["name"]}:{self._emoji["id"]}>'
    @emoji.setter
    def emoji(self, val: Union[Emoji, str, dict]):
        if val is None:
            self._emoji = None
        elif isinstance(val, str):
            self._emoji = {
                "name": val
            }
        elif isinstance(val, Emoji):
            self._emoji = {
                "id": val.id,
                "name": val.name,
                "animated": val.animated
            }
        elif isinstance(val, dict):
            self._emoji = val
        else:
            raise WrongType("emoji", val, ["str", "discord.Emoji", "dict"])

class Button(BaseButton, UseableComponent):
    """
    A ui-button

    Parameters
    ----------
    custom_id: :class:`str`
        A identifier for the button, max 100 characters
    label: :class:`str`, optional
        Text that appears on the button, max 80 characters; default \u200b ("empty" char)
    color: :class:`str` | :class:`int`, optional
        The color of the button; default "blurple"

        .. tip:

            You can either use a string for a color or an int. Color strings are: 
            (`primary`, `blurple`), (`secondary`, `grey`), (`succes`, `green`) and (`danger`, `Red`)
            
            If you want to use integers, take a lot at the :class:`~ButtonStyles` class

    emoji: :class:`discord.Emoji` | :class:`str`, optional
        The emoji displayed before the text; default MISSING
    new_line: :class:`bool`, optional
        Whether a new line should be added before the button; default False
    disabled: :class:`bool`, optional
        Whether the button is disabled; default False
    """
    def __init__(self, custom_id, label="\u200b", color = "blurple", emoji=None, new_line=False, disabled=False) -> None:
        """
        Creates a new ui-button

        Example:
        ```py
        Button("my_custom_id", "This is a cool button", "green", new_line=True)
        ```
        """
        BaseButton.__init__(self, label, color, emoji, new_line, disabled)
        UseableComponent.__init__(self, self.component_type)
        self.custom_id = custom_id
    def __repr__(self) -> str:
        return f"<discord_ui.Button({self.custom_id}:{self.content})>"
    def copy(self) -> 'Button':
        return Button(self.custom_id, self.label, self.color, self.emoji, self.new_line, self.disabled)

    @classmethod
    def _fromData(cls, data, new_line=False) -> 'Button':
        """
        Returns a new button initialized from api response data

        Returns
        -------
        Button
            The initialized button
        """
        return Button(data["custom_id"], data.get("label"), data["style"], data.get("emoji"), new_line, data.get("disabled", False))

class LinkButton(BaseButton):
    """
    A discord-ui linkbutton

    Parameters
    ----------
    url: :class:`str`
        A url which will be opened when pressing the button
    label: :class:`str`, optional
        Text that appears on the button, max 80 characters; default \u200b ("empty" char)
    emoji: :class:`discord.Emoji` | :class:`str`, optional
        Emoji that appears before the label; default MISSING
    new_line: :class:`bool`, optional
        Whether a new line should be added before the button; default False
    disabled: :class:`bool`, optional
        Whether the button is disabled; default False
    """
    def __init__(self, url, label="\u200b", emoji=None, new_line=False, disabled=False) -> None:
        """
        Creates a new LinkButton object
        
        Example:
        ```py
        LinkButton("https://discord.com/", "press me (if you can)!", emoji="😀", disabled=True)
        ```
        """
        BaseButton.__init__(self, label, ButtonStyles.url, emoji, new_line, disabled)
        self._url = None
        self.url = url

    def __repr__(self) -> str:
        return f"<discord_ui.LinkButton({self.custom_id}:{self.content})>"
    def copy(self) -> 'LinkButton':
        return LinkButton(self.url, self.label, self.emoji, self.new_line, self.disabled)

    @property
    def url(self) -> str:
        """
        The link which will be opened upon pressing the button

        :type: :class:`str`
        """
        return self._url
    @url.setter
    def url(self, val: str):
        if not isinstance(val, str):
            raise WrongType("url", val, "str")
        self._url = str(val)

    @classmethod
    def _fromData(cls, data, new_line=False) -> 'LinkButton':
        return LinkButton(data["url"], data.get("label"), data.get("emoji"), new_line, data.get("disabled", False))


class ButtonStyles(IntEnum):
    """
    A list of button styles (colors) in message components
    """
    Primary     =     	 blurple        = 1
    Secondary   =         grey          = 2
    Succes      =        green          = 3
    Danger      =         red           = 4
    url                                 = 5

    def __str__(self) -> str:
        return self.name

    @classmethod
    def getColor(cls, s):
        if isinstance(s, int):
            return cls(s)
        if isinstance(s, cls):
            return s
        s = s.lower()
        if s in ("blurple", "primary"):
            return cls.blurple
        if s in ("grey", "gray", "secondary"):
            return cls.grey
        if s in ("green", "succes"):
            return cls.green
        if s in ("red", "danger"):
            return cls.red
# endregion


class ActionRow():
    """
    Alternative to setting ``new_line`` in a full component list or putting the components in a list
        
    Only works for :class:`~Button` and :class:`~LinkButton`, because :class:`~SelectMenu` is always in a new line
    
    Parameters
    ----------
        disbaled: :class:`bool`, optional
            Whether all components should be disabled; default False   
    """
    def __init__(self, *items, disabled = False):
        """
        Creates a new component list

        Examples
        ```py
        ActionRow(Button(...), Button(...))
        
        ActionRow([Button(...), Button(...)])
        ```
        """
        self.items = items[0] if all(isinstance(i, list) for i in items) else items
        """The componetns in the action row"""
        self.component_type = 1
        self.disable(disabled)
        
    def disable(self, disable=True) -> 'ActionRow':
        for i, _ in enumerate(self.items):
            self.items[i].disabled = disable
        return self
    def filter(self, check = lambda x: ...):
        """
        Filters all components
        
        Parameters
        ----------
            check: :class:`lambda`
                What condition has to be True that the component will pass the filter
        
        Returns
        -------
            :returns: The filtered components
            :type: List[:class:`~Button` | :class:`~LinkButton`]
        
        """
        return [x for x in self.items if check(x)]


class ComponentType(IntEnum):
    """
    A list of component types
    """
    Action_row      =           1
    Button          =           2
    Select          =           3

    def __str__(self) -> str:
        return self.name

def make_component(data, new_line = False):
    if ComponentType(data["type"]) == ComponentType.Button:
        if data["style"] == ButtonStyles.url:
            return LinkButton._fromData(data, new_line)
        return Button._fromData(data, new_line)
    if ComponentType(data["type"]) is ComponentType.Select:
        return SelectMenu._fromData(data)
    # if data["type"] == ComponentType.ACTION_ROW:
        # return ActionRow._fromData(data)


class DuplicateComponentListener(Exception):
    def __init__(self, custom_id, *args: object) -> None:
        super().__init__("Listener for custom_id " + custom_id + " is already in use")
class InvalidLength(Exception):
    """This exception is thrown whenever a invalid length was provided"""
    def __init__(self, my_name, _min=None, _max=None, *args: object) -> None:
        if _min is not None and _max is not None:
            err = "Length of '" + my_name + "' must be between " + str(_min) + " and " + str(_max)
        elif _min is None and _max is not None:
            err = "Length of '" + my_name + "' must be less than " + str(_max)
        elif _min is not None and _max is None:
            err = "Lenght of '" + my_name + "' must be more than " + str(_min)
        super().__init__(err)
class OutOfValidRange(Exception):
    """This exception is thrown whenever a value was ot of its valid range"""
    def __init__(self, name, _min, _max, *args: object) -> None:
        super().__init__("'" + name + "' must be in range " + str(_min) + " and " + str(_max))
class WrongType(TypeError):
    """This exception is thrown whenever a value is of the wrong type"""
    def __init__(self, name, me, valid_type, *args: object) -> None:
        super().__init__("'" + name + "' must be of type " + (str(valid_type) if not isinstance(valid_type, list) else ' or '.join(valid_type)) + ", not " + str(type(me)))
class InvalidEvent(Exception):
    """This exception is thrown whenever a invalid eventname was passed"""
    def __init__(self, name, events, *args: object) -> None:
        super().__init__("Invalid event name, event must be " + " or ".join(events) + ", not " + str(name))
class MissingListenedComponentParameters(Exception):
    """This exception is thrown whenever a callback for a listening component is missing parameters"""
    def __init__(self, *args: object) -> None:
        super().__init__("Callback function for listening components needs to accept one parameter (the used component)", *args)
class CouldNotParse(Exception):
    """This exception is thrown whenever the libary was unable to parse the data with the given method"""
    def __init__(self, data, type, method, *args: object) -> None:
        super().__init__("Could not parse '" + str(data) + " [" + str(type) + "]' with method " + str(method), *args)


import asyncio
from .errors import WrongType
from .tools import MISSING, _or, components_to_dict, setup_logger

import discord
from discord.http import Route

import json
from typing import List

logging = setup_logger(__name__)

class BetterRoute(Route):
    BASE = "https://discord.com/api/v9"

async def send_files(route, files, payload, http):
    """Sends files"""

    form = []
    form.append({'name': 'payload_json', 'value': json.dumps(payload, separators=(',', ':'), ensure_ascii=True)})

    if len(files) == 1:
        file = files[0]
        form.append({
            'name': 'file',
            'value': file.fp,
            'filename': file.filename,
            'content_type': 'application/octet-stream'
        })
    else:
        for index, file in enumerate(files):
            form.append({
                'name': 'file%s' % index,
                'value': file.fp,
                'filename': file.filename,
                'content_type': 'application/octet-stream'
            })

    return await http.request(route, form=form, files=files)

def jsonifyMessage(content=MISSING, tts=False, embed: discord.Embed=MISSING, embeds: List[discord.Embed]=MISSING, attachments: List[discord.Attachment]=MISSING, nonce: int=MISSING,
                allowed_mentions: discord.AllowedMentions=MISSING, reference: discord.MessageReference=MISSING, mention_author: bool=MISSING, components: list=MISSING, stickers: List[discord.Sticker]=MISSING, suppress: bool=MISSING, flags=MISSING):
    """Turns parameters from send functions into a payload for requests"""
    
    payload = {"tts": tts}

    if content is not MISSING:
        if content is None:
            payload["content"] = ""
        else:
            payload["content"] = str(content)

    if suppress is not MISSING:
        flags = discord.MessageFlags._from_value(_or(flags, discord.MessageFlags.DEFAULT_VALUE))
        flags.suppress_embeds = suppress
        payload['flags'] = flags.value
    
    if nonce is not MISSING:
        payload["nonce"] = nonce
    
    if embed is not MISSING or embeds is not MISSING:
        if embed is not MISSING and embeds is MISSING:
            embeds = [embed]
        elif embed is not MISSING and embeds is not MISSING:
            embeds.append(embed)
        if embeds is None or embed is None:
            embeds = []
        elif not all(isinstance(x, discord.Embed) for x in embeds):
            raise WrongType("embeds", embeds, 'list[discord.Embed]')
        payload["embeds"] = [em.to_dict() for em in embeds]

    if attachments is not MISSING:
        if attachments is None:
            payload["attachments"] = []
        else:
            if not all(isinstance(x, discord.Attachment) for x in attachments):
                raise WrongType("attachments", attachments, "List[discord.attachment]")
            payload["attachments"] = [x.to_dict() for x in attachments]

    if reference is not MISSING:
        if not isinstance(reference, (discord.MessageReference, discord.Message)):
            raise WrongType("reference", reference, ['discord.MessageReference', 'discord.Message'])
        if isinstance(reference, discord.MessageReference):
            payload["message_reference"] = reference.to_dict()
        elif isinstance(reference, discord.Message):
            payload["message_reference"] = discord.MessageReference.from_message(reference).to_dict()

    if allowed_mentions is not MISSING:
        if allowed_mentions is None:
            payload["allowed_mentions"] = discord.AllowedMentions()
        else: 
            if not isinstance(allowed_mentions, discord.AllowedMentions):
                raise WrongType("allowed_mentions", allowed_mentions, "discord.AllowedMentions")
        payload["allowed_mentions"] = allowed_mentions.to_dict()
    if mention_author is not MISSING:
        allowed_mentions = payload["allowed_mentions"] if "allowed_mentions" in payload else discord.AllowedMentions().to_dict()
        allowed_mentions['replied_user'] = mention_author
        payload["allowed_mentions"] = allowed_mentions

    if components is not MISSING:
        if components is None:
            payload["components"] = []
        else:
            payload["components"] = components_to_dict(components)

    if stickers is not MISSING:
        if stickers is None:
            payload["stickers"] = []
        else:
            payload["stickers"] = [s.id for s in stickers]

    return payload

def handle_rate_limit(data):
    logging.error("You are being rate limited. Retrying after " + str(data["retry_after"]) + " seconds")
    return asyncio.sleep(data["retry_after"])


from .errors import InvalidEvent, OutOfValidRange, WrongType
from .slash.errors import AlreadyDeferred, EphemeralDeletion
from .slash.types import ContextCommand, OptionType, SlashCommand, SlashOption, SlashPermission, SlashSubcommand
from .tools import MISSING, setup_logger, _none
from .http import BetterRoute, jsonifyMessage, send_files
from .components import ActionRow, Button, Component, LinkButton, SelectMenu, SelectOption, UseableComponent, make_component

import discord
from discord.ext.commands import Bot
from discord.errors import HTTPException
from discord.state import ConnectionState

from typing import List, Union, Dict
try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

logging = setup_logger("discord-ui")


class InteractionType:
    PING                        =       Ping        =           1
    APPLICATION_COMMAND         =      Command      =           2
    MESSAGE_COMPONENT           =     Component     =           3

class Interaction():
    def __init__(self, state, data, user=None, message=None) -> None:
        self._state: ConnectionState = state

        self.deferred: bool = False
        self.responded: bool = False
        self._deferred_hidden: bool = False
        self._original_payload: dict = data

        self.author: Union[discord.Member, discord.User] = user
        """The user who created the interaction"""
        self.application_id: int = data["application_id"]
        """The ID of the bot application"""
        self.token: str = data["token"]
        """The token for responding to the interaction"""
        self.id: int = data["id"]
        """The id of the interaction"""
        self.type: int = data["type"]
        """The type of the interaction. See :class:`~InteractionType` for more information"""
        self.version: int = data["version"]
        self.data: dict = data["data"]
        """The passed data of the interaction"""
        self.channel_id: int = int(data.get("channel_id")) if data.get("channel_id") is not None else None
        """The channel-id where the interaction was created"""
        self.guild_id: int = int(data["guild_id"]) if data.get("guild_id") is not None else None
        """The guild-id where the interaction was created"""
        self.message: Message = message
        """The message of the interaction"""

    @property
    def guild(self) -> discord.Guild:
        """
        The guild where the interaction was created
        
        :type: :class:`discord.Guild`
        """
        return self._state._get_guild(self.guild_id)
    @property
    def channel(self) -> Union[discord.TextChannel, discord.DMChannel]:
        """
        The channel where the interaction was created
        
        :type: :class:`discord.TextChannel` | :class:`discord.DMChannel`
        """
        return self._state.get_channel(self.channel_id) or self._state.get_channel(self.author.id)

    async def defer(self, hidden=False):
        """
        This will acknowledge the interaction. This will show the (*Bot* is thinking...) Dialog

        .. note::
            
            This function should be used if the bot needs more than 15 seconds to respond
        
        Parameters
        ----------
            hidden: :class:`bool`, optional
                Whether the loading thing should be only visible to the user; default False.
        
        """
        if self.deferred:
            logging.error(AlreadyDeferred())
            return

        body = {"type": 5}
        
        if hidden is True:
            body["data"] = {"flags": 64}
            self._deferred_hidden = True
        
        await self._state.http.request(BetterRoute("POST", f'/interactions/{self.id}/{self.token}/callback'), json=body)
        self.deferred = True

    async def respond(self, content=MISSING, *, tts=False, embed=MISSING, embeds=MISSING, file=MISSING, files=MISSING, nonce=MISSING,
    allowed_mentions=MISSING, mention_author=MISSING, components=MISSING, delete_after=MISSING, listener=MISSING, 
    hidden=False, ninja_mode=False) -> Union['Message', 'EphemeralMessage']:
        """
        Responds to the interaction
        
        Parameters
        ----------
        content: :class:`str`, optional
            The raw message content
        tts: :class:`bool`
            Whether the message should be send with text-to-speech
        embed: :class:`discord.Embed`
            Embed rich content
        embeds: List[:class:`discord.Embed`]
            A list of embeds for the message
        file: :class:`discord.File`
            The file which will be attached to the message
        files: List[:class:`discord.File`]
            A list of files which will be attached to the message
        nonce: :class:`int`
            The nonce to use for sending this message
        allowed_mentions: :class:`discord.AllowedMentions`
            Controls the mentions being processed in this message
        mention_author: :class:`bool`
            Whether the author should be mentioned
        components: List[:class:`~Button` | :class:`~LinkButton` | :class:`~SelectMenu`]
            A list of message components to be included
        delete_after: :class:`float`
            After how many seconds the message should be deleted, only works for non-hiddend messages; default MISSING
        listener: :class:`Listener`
            A component-listener for this message
        hidden: :class:`bool`
            Whether the response should be visible only to the user 
        ninja_mode: :class:`bool`
            If true, the client will respond to the button interaction with almost nothing and returns nothing
        
        Returns
        -------
        :return: Returns the sent message
        :type: :class:`~Message` | :class:`EphemeralMessage`
        """
        if ninja_mode is True or all(y in [MISSING, False] for x, y in locals().items() if x not in ["self"]):
            try:
                route = BetterRoute("POST", f'/interactions/{self.id}/{self.token}/callback')
                r = await self._state.http.request(route, json={
                    "type": 6
                })
                return
            except HTTPException as x:
                if "value must be one of (4, 5)" in str(x).lower():
                    logging.warning(str(x) + "\n" + "The 'ninja_mode' parameter is not supported for slash commands!")
                    ninja_mode = False
                else:
                    raise x

        if self.responded is True:
            return await self.send(content=content, tts=tts, embed=embed, embeds=embeds, nonce=nonce, allowed_mentions=allowed_mentions, mention_author=mention_author, components=components, listener=listener, hidden=hidden)

        if components is MISSING and listener is not MISSING:
            components = listener.to_components()        
        payload = jsonifyMessage(content=content, tts=tts, embed=embed, embeds=embeds, nonce=nonce, allowed_mentions=allowed_mentions, mention_author=mention_author, components=components)
        
        if self._deferred_hidden is hidden:
            if self._deferred_hidden is False and hidden is True:
                logging.warning("Your response should be hidden, but the interaction was deferred public. This results in a public response.")
            if self._deferred_hidden is True and hidden is False:
                logging.warning("Your response should be public, but the interaction was deferred hidden. This results in a hidden response.")
        hide_message = self._deferred_hidden or not self.deferred and hidden is True


        r = None
        if not _none(delete_after) and hide_message is True:
            raise EphemeralDeletion()

        if hide_message:
            payload["flags"] = 64

        if (file is not MISSING or files is not MISSING) and self.deferred is False:
            await self.defer(hidden=hide_message)
        
        if self.deferred is False and hide_message is False:
            route = BetterRoute("POST", f'/interactions/{self.id}/{self.token}/callback')
            r = await self._state.http.request(route, json={
                "type": 4,
                "data": payload
            })
        else:
            if self.deferred is False and hide_message is True:
                await self.defer(hide_message)
            route = BetterRoute("PATCH", f'/webhooks/{self.application_id}/{self.token}/messages/@original')
            if file is not MISSING or files is not MISSING:
                r = await send_files(route=route, files=[file] if files is MISSING else files, payload=payload, http=self._state.http)
            else:
                r = await self._state.http.request(route, json=payload)
        self.responded = True
        
        if hide_message is True:
            msg = EphemeralMessage(state=self._state, channel=self._state.get_channel(int(r["channel_id"])), data=r, application_id=self.application_id, token=self.token)
            if listener is not MISSING:
                listener._start(msg)
            return msg

        
        if not hide_message:
            responseMSG = await self._state.http.request(BetterRoute("GET", f"/webhooks/{self.application_id}/{self.token}/messages/@original"))
            msg = await getMessage(self._state, data=responseMSG, response=False)
            if not _none(delete_after):
                await msg.delete(delete_after)
            if listener is not MISSING:
                listener._start(msg)
            return msg

    async def send(self, content=None, *, tts=False, embed=MISSING, embeds=MISSING, file=MISSING, files=MISSING, nonce=MISSING,
    allowed_mentions=MISSING, mention_author=MISSING, components=MISSING, listener=MISSING, hidden=False) -> Union['Message', 'EphemeralMessage']:
        """
        Sends a message to the interaction using a webhook
        
        Parameters
        ----------
        content: :class:`str`, optional
            The raw message content
        tts: :class:`bool` 
            Whether the message should be send with text-to-speech
        embed: :class:`discord.Embed`
            Embed rich content
        embeds: List[:class:`discord.Embed`]
            A list of embeds for the message
        file: :class:`discord.File`
            The file which will be attached to the message
        files: List[:class:`discord.File`]
            A list of files which will be attached to the message
        nonce: :class:`int`
            The nonce to use for sending this message
        allowed_mentions: :class:`discord.AllowedMentions`
            Controls the mentions being processed in this message
        mention_author: :class:`bool`
            Whether the author should be mentioned
        components: List[:class:`~Button` | :class:`~LinkButton` | :class:`~SelectMenu`]
            A list of message components to be included
        listener: :class:`Listener`
            A component-listener for this message
        hidden: :class:`bool`
            Whether the response should be visible only to the user 
        ninja_mode: :class:`bool`
            If true, the client will respond to the button interaction with almost nothing and returns nothing
        
        Returns
        -------
        :return: Returns the sent message
        :type: :class:`~Message` | :class:`EphemeralMessage`
        """
        if self.responded is False:
            return await self.respond(content=content, tts=tts, embed=embed, embeds=embeds, file=file, files=files, nonce=nonce, allowed_mentions=allowed_mentions, mention_author=mention_author, components=components, listener=listener, hidden=hidden)

        if components is MISSING and listener is not MISSING:
            components = listener.to_components()
        payload = jsonifyMessage(content=content, tts=tts, embed=embed, embeds=embeds, nonce=nonce, allowed_mentions=allowed_mentions, mention_author=mention_author, components=components)
        
        if hidden:
            payload["flags"] = 64

        route = BetterRoute("POST", f'/webhooks/{self.application_id}/{self.token}')
        
        if file is not MISSING or files is not MISSING:
            r = await send_files(route=route, files=[file] if files is MISSING else files, payload=payload, http=self._state.http)
        else:
            r = await self._state.http.request(route, json=payload)

        if hidden is True:
            msg = EphemeralMessage(state=self._state, channel=self._state.get_channel(r["channel_id"]), data=r, application_id=self.application_id, token=self.token)
            if listener is not MISSING:
                listener._start(msg)
        msg = await getMessage(self._state, r, response=False)
        
        if listener is not MISSING:
            listener._start(msg)
        return msg
    def _handle_auto_defer(self, auto_defer):
        self.deferred = auto_defer[0]
        self._deferred_hidden = auto_defer[1]

class ComponentContext(Interaction, UseableComponent):
    """A received component"""
    def __init__(self, state, data, user, message) -> None:
        Interaction.__init__(self, state, data, user=user, message=message)
        UseableComponent.__init__(self, data["data"]["component_type"])

class SelectedMenu(Interaction, SelectMenu):
    """A :class:`~SelectMenu` object in which an item was selected"""
    def __init__(self, data, user, s, msg, client) -> None:
        Interaction.__init__(self, client._connection, data, user, msg)
        default = [i for i, o in enumerate(s.options) if o.default is True]
        SelectMenu.__init__(self, s.custom_id, s.options, s.min_values, s.max_values, s.placeholder, default[0] if len(default) == 1 else None, s.disabled)
        
        self.bot: Bot = client
        self.selected_options: List[SelectOption] = []
        """The list of the selected options"""
        self.selected_values: List[str] = []
        """The list of raw values which were selected"""
        for val in data["data"]["values"]:
            for x in self.options:
                if x.value == val:
                    self.selected_options.append(x)
                    self.selected_values.append(x.value)
        self.author: discord.Member = user
        """The user who selected the value"""

class PressedButton(Interaction, Button):
    """A :class:`~Button` object that was pressed"""
    def __init__(self, data, user, b, message, client) -> None:
        Interaction.__init__(self, client._connection, data, user, message)
        Button.__init__(self, b.custom_id, b.label, b.color, b.emoji, b.new_line, b.disabled)

        self._json = b.to_dict()
        self.bot: Bot = client
        self.author: discord.Member = user
        """The user who pressed the button"""

class SlashedCommand(Interaction, SlashCommand):
    """A :class:`~SlashCommand` object that was used"""
    def __init__(self, client, command: SlashCommand, data, user, args = None) -> None:
        Interaction.__init__(self, client._connection, data, user)
        SlashCommand.__init__(self, command.callback, command.name, command.description, command.options, guild_ids=command.guild_ids, guild_permissions=command.guild_permissions)
        for x in self.__slots__:
            setattr(self, x, getattr(command, x))

        self.bot: Bot = client
        self._json = command.to_dict()
        self.author: discord.Member = user
        """The user who used the command"""
        self.args: Dict[str, Union[str, int, bool, discord.Member, discord.TextChannel, discord.Role, float]] = args
        """The options that were received"""
        self.permissions: SlashPermission = command.guild_permissions.get(self.guild_id) if command.guild_permissions is not None else None
        """The permissions for this guild"""
class SlashedSubCommand(SlashedCommand, SlashSubcommand):
    """A Sub-:class:`~SlashCommand` command that was used"""
    def __init__(self, client, command, data, user, args = None) -> None:
        SlashedCommand.__init__(self, client, command, data, user, args)
        SlashSubcommand.__init__(self, command.callback, command.base_names, command.name, command.description, command.options, command.guild_ids, command.default_permission, command.guild_permissions)
            

class SlashedContext(Interaction, ContextCommand):
    def __init__(self, client, command: ContextCommand, data, user, param) -> None:
        Interaction.__init__(self, client._connection, data, user)
        ContextCommand.__init__(self, command.command_type, command.callback, command.name, guild_ids=command.guild_ids, guild_permissions=command.guild_permissions)
        for x in self.__slots__:
            setattr(self, x, getattr(command, x))
        
        self._json = command.to_dict()
        self.bot: Bot = client
        self.param: Union[Message, discord.Member, discord.User] = param
        """The parameter that was received"""
        self.permissions: SlashPermission = command.guild_permissions.get(self.guild_id) if command.guild_permissions is not None else None 
        """The permissions for this guild"""
        


async def getMessage(state: ConnectionState, data, response=True):
    """
    Async function to get the response message

    Parameters
    -----------------

    state: :class:`discord.state.ConnectionState`
        The discord bot client
    data: :class:`dict`
        The raw data
    user: :class:`discord.User`
        The User which pressed the button
    response: :class:`bool`
        Whether the Message returned should be of type `ResponseMessage` or `Message`

    Returns
    -------
    :class:`~Message` | :class:`~EphemeralMessage`
        The sent message
    """
    msg_base = data.get("message", data)

    channel = state.get_channel(int(data["channel_id"])) or state.get_channel(int(msg_base["author"]["id"]))
    if response:
        if msg_base["flags"] == 64:
            return EphemeralResponseMessage(state=state, channel=channel, data=data.get("message", data))
        return Message(state=state, channel=channel, data=msg_base)

    if msg_base["flags"] == 64:
        return EphemeralMessage(state=state, channel=channel, data=msg_base)
    return Message(state=state, channel=channel, data=msg_base)

class Message(discord.Message):
    """A fixed :class:`discord.Message` optimized for components"""
    def __init__(self, *, state, channel, data):
        self.__slots__ = discord.Message.__slots__ + ("components", "supressed")
        self._payload = data

        self._state: ConnectionState = None
        discord.Message.__init__(self, state=state, channel=channel, data=data)
        self.components: List[Union[Button, LinkButton, SelectMenu]] = []
        """The components in the message
        
        :type: List[]:class:`~Button` | :class:`~LinkButton` | :class:`SelectMenu`]
        """
        self.suppressed = False
        self._update_components(data)

    # region attributes
    @property
    def buttons(self) -> List[Union[Button, LinkButton]]:
        """The button components in the message
        
        :type: List[:class:`~Button` | :class:`~LinkButton`]
        """
        if hasattr(self, "components") and self.components is not None:
            return [x for x in self.components if isinstance(x, (Button, LinkButton))]
        return []
    @property
    def select_menus(self) -> List[SelectMenu]:
        """The select menus components in the message

        :type: List[:class:`~SelectMenu`]
        """
        if hasattr(self, "components") and self.components is not None:
            return [x for x in self.components if isinstance(x, SelectMenu)]
        return []
    @property
    def action_row(self) -> ActionRow:
        return ActionRow(self.components)
    # endregion

    def _update_components(self, data):
        """Updates the message components"""
        if data.get("components") is None:
            self.components = []
            return
        self.components = []
        if len(data["components"]) == 0:
            pass
        elif len(data["components"]) > 1:
            # multiple lines
            for componentWrapper in data["components"]:
                # newline
                for index, com in enumerate(componentWrapper["components"]):
                    self.components.append(make_component(com, index==0))
        elif len(data["components"][0]["components"]) > 1:
            # All inline
            for index, com in enumerate(data["components"][0]["components"]):
                self.components.append(make_component(com, index==0))
        else:
            # One button
            component = data["components"][0]["components"][0]
            self.components.append(make_component(component))

    def _update(self, data):
        super()._update(data)
        self._update_components(data)

    async def edit(self, content=MISSING, *, embed=MISSING, embeds=MISSING, attachments=MISSING, suppress=MISSING, 
        delete_after=MISSING, allowed_mentions=MISSING, components=MISSING):
        """Edits the message and updates its properties

        .. note::

            If a paremeter is `None`, the attribute will be removed from the message

        Parameters
        ----------------
        content: :class:`str`
            The new message content
        embed: :class:`discord.Embed`
            The new embed of the message
        embeds: List[:class:`discord.Embed`]
            The new list of discord embeds
        attachments: List[:class:`discord.Attachment`]
            A list of new attachments
        supress: :class:`bool`
            Whether the embeds should be shown
        delete_after: :class:`float`
            After how many seconds the message should be deleted
        allowed_mentions: :class:`discord.AllowedMentions`
            The mentions proceeded in the message
        components: List[:class:`~Button` | :class:`~LinkButton` | :class:`~SelectMenu`]
            A list of components to be included the message
        """
        payload = jsonifyMessage(content, embed=embed, embeds=embeds, allowed_mentions=allowed_mentions, attachments=attachments, suppress=suppress, flags=self.flags.value, components=components)
        data = await self._state.http.edit_message(self.channel.id, self.id, **payload)
        self._update(data)

        if not _none(delete_after):
            await self.delete(delay=delete_after)

    async def disable_action_row(self, row, disable = True):
        """Disables an action row of components in the message
        
        Parameters
        ----------
            row: :class:`int` |  :class:`range`
                Which rows to disable, first row is ``0``; If row parameter is of type :class:`int`, the nth row will be disabled, if type is :class:`range`, the range is going to be iterated and all rows in the range will be disabled

            disable: :class:`bool`, optional
                Whether to disable (``True``) or enable (``False``) the components; default True

        Raises
        ------
            :raises: :class:`discord_ui.errors.OutOfValidRange` : The specified range was out of the possible range of the component rows 
            :raises: :class:`discord_ui.errors.OutOfValidRange` : The specified row was out of the possible range of the component rows
        
        """
        comps = []
        if isinstance(row, range):
            for i, _ in enumerate(self.action_rows):
                if i >= len(self.action_rows) - 1 or i < 0:
                    raise OutOfValidRange("row[" + str(i) + "]", 0, len(self.action_rows) - 1)
                for comp in self.action_rows[i]:
                    if i in row:
                        comp.disabled = disable
                    comps.append(comp)
        else:
            for i, _ in enumerate(self.action_rows):
                if i >= len(self.action_rows) - 1 or i < 0:
                    raise OutOfValidRange("row", 0, len(self.action_rows) - 1)
                for comp in self.action_rows[i]:
                    if i == row:
                        comp.disabled = disable
                    comps.append(comp)
        await self.edit(components=comps)

    async def disable_components(self, disable = True):
        """Disables all component in the message
        
        Parameters
        ----------
            disable: :class:`bool`, optional
                Whether to disable (``True``) or enable (``False``) als components; default True
        
        """
        fixed = []
        for x in self.components:
            x.disabled = disable
            fixed.append(x)
        await self.edit(components=fixed)

    @property
    def action_rows(self):
        """The action rows in the message

        :type: List[:class:`~Button` | :class:`LinkButton` | :class:`SelectMenu`]
        """
        rows: List[List[Union[Button, LinkButton, SelectMenu]]] = []

        c_row = []
        i = 0
        for x in self.components:
            if getattr(x, 'new_line', True) == True and i > 0:
                rows.append(ActionRow(c_row))
                c_row = []
            c_row.append(x)
            i += 1
        if len(c_row) > 0:
            rows.append(c_row) 
        return rows

    async def wait_for(self, event_name: Literal["select", "button", "component"], client, custom_id=None, by=None, check=lambda component: True, timeout=None) -> Union[PressedButton, SelectedMenu, ComponentContext]:
        """Waits for a message component to be invoked in this message

        Parameters
        -----------
            event_name: :class:`str`
                The name of the event which will be awaited [``"select"`` | ``"button"`` | ``"component"``]
                
                .. note::

                    ``event_name`` must be ``select`` for a select menu selection, ``button`` for a button press and ``component`` for any component

            client: :class:`discord.ext.commands.Bot`
                The discord client
            custom_id: :class:`str`, Optional
                Filters the waiting for a custom_id
            by: :class:`discord.User` | :class:`discord.Member` | :class:`int` | :class:`str`, Optional
                The user or the user id by that has to create the component interaction
            check: :class:`function`, Optional
                A check that has to return True in order to break from the event and return the received component
                    The function takes the received component as the parameter
            timeout: :class:`float`, Optional
                After how many seconds the waiting should be canceled. 
                Throws an :class:`asyncio.TimeoutError` Exception

        Raises
        ------
            :raises: :class:`discord_ui.errors.InvalidEvent` : The event name passed was invalid 

        Returns
        ----------
            :returns: The component that was waited for
            :type: :class:`~PressedButton` | :class:`~SelectedMenu`

        Example
        ---------

        .. code-block::

            # send a message with comoponents
            msg = await ctx.send("okayy", components=[Button("a_custom_id", ...)])
            try:
                # wait for the button
                btn = await msg.wait_for("button", client, "a_custom_id", by=ctx.author, timeout=20)
                # send response
                btn.respond()
            except asyncio.TimeoutError:
                # no button press was received in 20 seconds timespan
        """
        def _check(com):
            if com.message.id == self.id:
                statements = []
                if not _none(custom_id):
                    statements.append(com.custom_id == custom_id)
                if not _none(by):
                    statements.append(com.author.id == (by.id if hasattr(by, "id") else int(by)))
                if not _none(check):
                    statements.append(check(com))
                return all(statements)
            return False
        if not isinstance(client, Bot):
            raise WrongType("client", client, "discord.ext.commands.Bot")
        
        if event_name.lower() == "button":
            return await client.wait_for('button_press', check=_check, timeout=timeout)
        if event_name.lower() == "select":
            return await client.wait_for("menu_select", check=_check, timeout=timeout)
        if event_name.lower() == "component":
            return await client.wait_for("component", check=_check, timeout=timeout)
        
        raise InvalidEvent(event_name, ["button", "select", "component"])

    async def put_listener(self, listener):
        """Adds a listener to this message and edits it if the components are missing
        
        Parameters
        ----------
        listener: :class:`Listener`
            The listener which should be put to the message
        
        """
        if len(self.components) == 0:
            await self.edit(components=listener.to_components())
        self.attach_listener(listener)
        
    def attach_listener(self, listener):
        """Attaches a listener to this message after it was sent
        
        Parameters
        ----------
        listener: :class:`Listener`
            The listener that should be attached
        
        """
        listener._start(self)
    def remove_listener(self):
        """Removes the listener from this message"""
        try:
            del self._state._component_listeners[str(self.id)]
        except KeyError:
            pass


class WebhookMessage(Message, discord.WebhookMessage):
    def __init__(self, *, state, channel, data):
        Message.__init__(self, state=state, channel=channel, data=data)
        discord.WebhookMessage.__init__(self, state=state, channel=channel, data=data)
    async def edit(self, *args, **fields):
        """Edits the message

        content: :class:`str`, Optional
            The content to edit the message with or None to clear it.
        embed: :class:`discord.Embed`
            Embed rich content
        embeds: List[:class:`discord.Embed`], Optional
            A list of embeds to edit the message with.
        supress: :class:`bool`
            Whether to supress all embeds in the message
        allowed_mentions: :class`discord.AllowedMentions`
            Controls the mentions being processed in this message. See `discord.abc.Messageable.send` for more information.
        """
        return await self._state._webhook._adapter.edit_webhook_message(message_id=self.id, payload=jsonifyMessage(*args, **fields))
    

class EphemeralMessage(Message):
    """Represents a hidden (ephemeral) message"""
    def __init__(self, state, channel, data, application_id=None, token=None):
        #region fix reference keyerror
        if data.get("message_reference"):
            if data["message_reference"].get("channel_id") is None:
                data["message_reference"]["channel_id"] = str(channel.id)
        #endregion
        Message.__init__(self, state=state, channel=channel, data=data)
        self._application_id = application_id
        self._interaction_token = token
    async def edit(self, *args, **fields):
        r = BetterRoute("PATCH", f"/webhooks/{self._application_id}/{self._interaction_token}/messages/{self.id}")
        self._update(await self._state.http.request(r, json=jsonifyMessage(*args, **fields)))        
    async def delete(self):
        """Override for delete function that will throw an exception"""
        raise EphemeralDeletion()

class EphemeralResponseMessage(Message):
    """A ephemeral message wich was created from an interaction
    
    .. important::

        Methods like `.edit()`, which change the original message, need a `token` paremeter passed in order to work
    """
    def __init__(self, *, state, channel, data):
        Message.__init__(self, state=state, channel=channel, data=data)

    async def edit(self, token, *args, **fields):
        """Edits the message
        
        Parameters
        ----------
            token: :class:`str`
                The token of the interaction with wich this ephemeral message was sent
            fields: :class:`kwargs`
                The fields to edit (ex. `content="...", embed=..., attachments=[...]`)

            Example

            .. code-block::

                async def testing(ctx):
                    msg = await ctx.send("hello hidden world", components=[Button("test")])
                    btn = await msg.wait_for("button", client)
                    await btn.message.edit(ctx.token, content="edited", components=None)
        
        """
        route = BetterRoute("PATCH", f"/webhooks/{self.interaction.application_id}/{token}/messages/{self.id}")
        self._update(await self._state.http.request(route, json=jsonifyMessage(*args, **fields)))
    async def delete(self):
        """Override for delete function that will throw an exception"""
        raise EphemeralDeletion()
    async def disable_components(self, token, disable = True):
        """Disables all component in the message
        
        Parameters
        ----------
            disable: :class:`bool`, optional
                Whether to disable (``True``) or enable (``False``) als components; default True
        
        """
        fixed = []
        for x in self.components:
            x.disabled = disable
            fixed.append(x)
        await self.edit(token, components=fixed)



"""
discord_ui.listeners
=====================

A module for listeners that are going to be attached to a message. 
You could think of it like a cog but for message components.


.. code-block::

    from discord_ui import Listener


- - -

Usage
======

To use the ``Listener`` class, you need to create a subclass of it

Example

.. code-block::

    class MyListener(Listener):
        ...

Listeners
---------

To add a button listener, you need to use the ``Listener.button`` deorator

.. code-block::

    class MyListener(Listener):
        ...

        @Listener.button("custom_id here")
        async def somebutton(self, ctx):
            ...

This will add a button listener that will wait for a button with the custom id "custom_id here"


To add a select listener, you need to use the ``Listener.select`` decoratorr

.. code-block::

    class MyListener(Listener):
        ...

        @Listener.select("my_id")
        async def someselect(self, ctx):
            ...
    
This will add a select menu listener that will wait for a select menu with the custom id "my_id"

You can filter the select callback with the ``values`` parameter

.. code-block::

    class MyListener(Listener):
        ...

        @Listener.select("my_id", values=["2"])
        async def someselect(self, ctx):
            ...

The callback will now be only called if the selected values of a select menu equal to ``values``

There are some more usefull things you can use


Class
-----

You can add a timeout to the listener after which the listener will be removed from the message

.. code-block::

    .. code-block::

    class MyListener(Listener):
        def __init__(self):
            self.timeout = 20   # 20 seconds timeout

If you **set** timeout to ``None``, the listener will never timeout

Sending
--------

To send components and add the listener, you can use five different ways

First method:

.. code-block::

    @bot.listen()
    async def on_message(message)
        class MyListener(Listener):
            def __init__(self):
                pass
            @Listener.button("test")
            async def test(self, ctx):
                ...

        await message.channel.send("showcase", components=[Button("test", "this is a showcase")], listener=MyListener())

Second method:

.. code-block::

    @bot.listen()
    async def on_message(message)
        class MyListener(Listener):
            def __init__(self):
                self.components = [Button("test", "this is a showcase")]
            @Listener.button("test")
            async def test(self, ctx):
                ...

        # MyListener.components will be used for the components in the message
        await message.channel.send("showcase", listener=MyListener())

Third method:

.. code-block::

    @bot.listen()
    async def on_message(message)
        class MyListener(Listener):
            def __init__(self):
                ...
            
            @Listener.button("test")
            async def test(self, ctx):
                ...

        msg = await message.channel.send("showcase", components=[Button("test", "this is a showcase")])
        msg.attach_listener(MyListener())

Fourth method:

.. code-block::

    ui = discord_ui.UI(bot)

    @bot.listen()
    async def on_message(message)
        class MyListener(Listener):
            def __init__(self):
                pass
            @Listener.button("test")
            async def test(self, ctx):
                ...

        msg = await message.channel.send("showcase", components=[Button("test", "this is a showcase")])
        ui.components.attach_listener_to(msg, MyListener())

And the last method:

.. code-block::

    ui = discord_ui.UI(bot)

    @bot.listen()
    async def on_message(message)
        class MyListener(Listener):
            def __init__(self):
                pass
            @Listener.button("test")
            async def test(self, ctx):
                ...

        msg = await message.channel.send("showcase", components=[Button("test", "this is a showcase")])
        MyListener.attach_me_to(msg)
"""

from .tools import setup_logger
from .components import Button, ComponentType, LinkButton, SelectMenu
from .receive import ComponentContext, Message

import discord
from discord.ext import commands
from discord.state import ConnectionState

import asyncio
from inspect import getmembers
from typing import Dict, List, Union

logging = setup_logger(__name__)

class _Listener():
    def __init__(self, callback, custom_id, component_type, values=None) -> None:
        self.callback = callback
        self.custom_id = custom_id
        self.type = component_type
        self.target_values = [str(v) for v in values] if values is not None else None 

        self.__commands_checks__ = []
        if hasattr(self.callback, "__command_checks__"):
            self.__commands_checks__ = self.callback.__commands_checks__

    async def __call__(self, *args, **kwargs):
        return await self.invoke(*args, **kwargs)
    def add_check(self, check):
        self.__commands_checks__.append(check)
    def remove_check(self, check):
        self.__commands_checks__.remove(check)
    @property
    def checks(self):
        return self.__commands_checks__
    async def can_run(self, ctx):
        """Whether the command can be run"""
        predicates = self.checks
        if not predicates:
            # since we have no checks, then we just return True.
            return True
        return await discord.utils.async_all(predicate(ctx) for predicate in predicates)
    
    async def invoke(self, ctx, listener):
        if not await self.can_run(ctx):
            raise commands.errors.CheckFailure()
        await self.callback(listener, ctx)

class NoListenerFound(Exception):
    """Exception that will be thrown when no matching listener was found"""
    def __init__(self, msg=None, *args: object) -> None:
        super().__init__(msg or "Could not find a matching listener", *args)
class WrongUser(Exception):
    def __init__(self, msg=None, *args: object) -> None:
        super().__init__(msg or "Wrong user used component", *args)


class Listener():
    """A class for a listener attached to a message that will receive components of it.
        To use this class you have to create a subclass inhering this class


    Example
    -------

    .. code-block::

        class MyListener(Listener)
            ...

    Parameters
    -----------
    timeout: :class:`float`, optional
        A timeout after how many seconds the listener shouold be deleted.
        If ``None``, the listener will never timeout
    target_user: :class:`discord.Member` | :class:`discord.User`
        The user from which the interactions should be received. 
        Every interaction by other users will be ignored
    """
    def __init__(self, timeout=180.0, target_user=None) -> None:
        self.timeout: float = timeout
        """Timeout after how many seconds the listener should timeout and be deleted"""
        self.target_user: Union[discord.Member, discord.User] = target_user
        """The user from which the interaction has to come"""
        self.components: List[Union[List[Button, LinkButton, SelectMenu], Button, LinkButton, SelectMenu]] = []
        """The components that are going to be send together with the listener"""
        self.message: Message = None
        """The target message"""
    def __init_subclass__(cls) -> None:
        cls.__listeners__ = []

    @staticmethod
    def button(custom_id=None):
        """A decorator that will setup a callback for a button
        
        Parameters
        ----------
        custom_id: :class:`str`, optional
            The custom id of the target button.  If no custom_id is passed, the name of the callback will be used for the custom id.
                Note that you can't have two callbacks with the same function name

        Example
        --------

        .. code-block::

            class MyListener(Listener):
                def __init__(self, ...):
                    ...

                @Listener.button("my_id")
                async def callback(self, ctx):
                    ...
        
        """
        def wraper(callback):
            return _Listener(callback, custom_id or callback.__name__, ComponentType.Button)
        return wraper
    @staticmethod
    def select(custom_id=None, values=None):
        """A decorator that will set a callback up for a select menu
        
        Parameters
        ----------
        custom_id: :class:`str`, optional
            The custom id of the target menu. If no id specified, the name of the callback will be used.
                Note that you can't have two callbacks with the same function name
        values: List[:class:`str`], optional
            What values must be selected in order to invoke the callback; default None
        
        Example
        --------

        .. code-block::

            class MyListener(Listener):
            def __init__(self, ...)
                ...

            @Listener.select("my_id")
            async def callback(self, ctx):
                ...

            # This callback will be only called if the selected values of the menu are '2' 
            @Listener.select("my_id", values=['2'])
            async def othere_callback(self, ctx):
                ...
        """
        def wraper(callback):
            return _Listener(callback, custom_id or callback.__name__, ComponentType.Select, values)
        return wraper

    async def _call_listeners(self, interaction_component):
        listeners = self._get_listeners_for(interaction_component)
        if len(listeners) > 0:
            for listener in listeners:
                await listener.invoke(interaction_component, self)
        else:
            raise NoListenerFound()
    def _get_listeners(self) -> Dict[str, List[_Listener]]:
        all_listeners = [x[1] for x in getmembers(self, predicate=lambda x: isinstance(x, _Listener))]
        listeners = {}
        for lister in all_listeners:
            if not listeners.get(lister.custom_id):
                listeners[lister.custom_id] = []
            listeners[lister.custom_id].append(lister)
        return listeners
    def _get_listeners_for(self, interaction_component: ComponentContext) -> _Listener:
        listeners = self._get_listeners()
        listers = []
        for listener in listeners.get(interaction_component.custom_id, []):
            if hasattr(self, 'target_user') and self.target_user.id != interaction_component.author.id:
                continue
            if listener.type == interaction_component.component_type:
                if listener.target_values is not None:
                    if sorted(interaction_component.data["values"]) == sorted(listener.target_values):
                    # if all(v in interaction_component.data["values"] for v in listener.target_values):
                        listers.append(listener)
                else:
                    listers.append(listener)
        return listers
   
    def to_components(self):
        return self.components

    def _stop(self):
        del self._state._component_listeners[self._target_message_id]
        logging.debug("deleted listener")
    def _start(self, message):
        self.message = message
        self._state: ConnectionState = message._state
        self._target_message_id = str(self.message.id)
        self._state._component_listeners[self._target_message_id] = self
        
        # region removal
        loop = asyncio.get_event_loop()
        if getattr(self, 'timeout', None) is not None:
                loop.call_later(self.timeout, self._stop)
    def attach_me_to(self, message):
        """Attaches this listener to a message after it was sent
        
        Parameters
        ----------
        message: :class:`~Message`
            The target message
        
        """
        self._start(message)
    async def put_me_to(self, message):
        """Attaches this listener to a message and edits it if the message is missing components
        
        Parameters
        ----------
        message: :class:`~Message`
            The target message
        
        """
        if len(message.components) == 0:
            await message.edit(components=self.to_components())
        self.attach_me_to(message)


import logging
from typing import Any, List

class _MISSING:
    def __repr__(self) -> str:
        return "..."
    def __eq__(self, o: object) -> bool:
        return isinstance(o, _MISSING)
    def __ne__(self, o: object) -> bool:
        return not self.__eq__(o)
    def __str__(self) -> str:
        return self.__repr__()
    def __sizeof__(self) -> int:
        return 0
    def get(self, *args):
        return self

MISSING = _MISSING()


def _none(*args, empty_array=False):
    return all(x in [None, MISSING] + [[], [[]]][empty_array is True] for x in args)
def _or(*args, default=None):
    for i, _ in enumerate(args):
        if not _none(args[i]):
            return args[i]
    return default
def _default(default, *args, empty_array=True):
    if _none(*args, empty_array=empty_array):
        return default
    if len(args) == 1:
        return args[0]
    return args


def get_index(l: list, elem: Any, mapping = lambda x: x, default: int = -1) -> int:
    """Returns the index of an element in the list
    
    Parameters
    ----------
        l: :class:`List`
            The list to get from
        elem: :class:`Any`
            The element that has to be found
        mapping: :class:`function`
            A function that will be applied to the current element that is checked, before comparing it to the target
        default: :class:`Any`
            The element that will be returned if nothing is found; default -1

    Returns
    -------
        :returns: The found element
        :type: :class:`Any`
    

    Example:
    ```py
    any_list = [(1, 2), (2, 3), (3, 4)]
    get(any_list, 4, lambda x: x[1])
    ```
    """
    i = 0
    for x in l: 
        if mapping(x) == elem: 
            return i
        i += 1
    return default

def get(l: list, elem: Any, mapping = lambda x: x, default: Any = None, check=lambda x: True):
    """Gets a element from a list
    
    Parameters
    ----------
        l: :class:`List`
            The list to get from
        elem: :class:`Any`
            The element that has to be found
        mapping: :class:`function`
            A function that will be applied to the current element that is checked, before comparing it to the target
        default: :class:`Any`
            The element that will be returned if nothing is found; default None

    Returns
    -------
        :returns: The found element
        :type: :class:`Any`


    Example:
    ```py
    any_list = [(1, 2), (2, 3), (3, 4)]
    get(any_list, 4, lambda x: x[1])
    ```
    """
    for x in l:
        if mapping(x) == elem and check(x) is True:
            return x
    return default

def components_to_dict(*components) -> List[dict]:
    """Converts a list of components to a dict that can be used for other extensions
    
    Parameters
    ----------
    components: :class:`*args` | :class:`list`
        A list of components that should be converted.
        
        Example
        
        .. code-block::

            # List of components with component rows (everything in [] defines that it will be in it's own component row)
            components_to_dict(Button(...), [Button(...), Button(...)], SelectMenu(...), LinkButton)
            # or
            components_to_dict([Button(...), [LinkButton(...), Button(...)]])

    Raises
    ------
        :raises: :class:`Exception` : Invalid Data was passed
    
    Returns
    -------
        :returns: The converted data
        :type: List[:class:`dict`]


    """
    wrappers: List[List[Any]] = []
    component_list = []
    if len(components) == 1 and isinstance(components[0], list):
        components = components[0]

    if len(components) > 1:
        curWrapper = []
        i = 0
        for component in components:
            if hasattr(component, "items") or isinstance(component, list):
                if i > 0 and len(curWrapper) > 0:
                    wrappers.append(curWrapper)
                curWrapper = []
                wrappers.append(component.items if hasattr(component, "items") else component)
                continue
                
            # i > 0 => Preventing empty component field when first button wants to newLine 
            if component.component_type == 3:
                if i > 0:
                    wrappers.append(curWrapper)
                curWrapper = []
                wrappers.append(component)
            elif component.component_type == 2:
                if component.new_line and i > 0:
                    wrappers.append(curWrapper)
                    curWrapper = [component]
                else: 
                    curWrapper.append(component)
                    i += 1
        if len(curWrapper) > 0:
            wrappers.append(curWrapper)
    else:
        wrappers = [components]

    for wrap in wrappers:
        if isinstance(wrap, list) and not all(hasattr(x, "to_dict") for x in wrap):
            raise Exception("Components with types [" + ', '.join([str(type(x)) for x in wrap]) + "] are missing to_dict() method")
        component_list.append({"type": 1, "components": [x.to_dict() for x in wrap] if type(wrap) is list else [wrap.to_dict()]})
    return component_list

def setup_logger(name):
    """
    Thx redstone ;)
    https://github.com/RedstoneZockt/rotstein-dc-py/blob/main/rotstein_py/logging.py
    """

    logger = logging.getLogger(name)
    logger.setLevel(logging.ERROR)
    
    stream_handler = logging.StreamHandler()
    formatter = logging.Formatter(f"%(levelname)s: %(message)s")
    stream_handler.setFormatter(formatter)
    stream_handler.setLevel(logging.ERROR)
    
    logger.addHandler(stream_handler)
    return logger


"""
discord_ui.slash
~~~~~~~~~~~~~~~~~~~

The slash module for this libary

- - -

Here is everything about slashcommmands.
You shouldn't need to import this module unless you really now what you're doing. 

The only thing you might wanna import is `.ext`, but you can import that from `discord_ui.ext`.
Everything else here could only be needed for intellisense.

- - -

[Github](https://github.com/discord-py-ui/discord-ui/tree/main/discord_ui/slash)
"""

from . import ext
from .types import SlashCommand, SlashSubcommand, SlashOption, ContextCommand, MessageCommand, UserCommand, SlashPermission, OptionType, CommandType


from .cogs import BaseCallable, CogCommand, CogMessageCommand, CogSubCommandGroup, ListeningComponent
from .components import Component, ComponentType
from .slash.errors import NoAsyncCallback
from .errors import MissingListenedComponentParameters, WrongType
from .slash.tools import ParseMethod, cache_data, format_name, handle_options, handle_thing
from .slash.http import create_global_command, create_guild_command, delete_global_command, delete_guild_command, delete_guild_commands, edit_global_command, edit_guild_command, get_command, get_command_permissions, get_global_commands, get_guild_commands, delete_global_commands, get_id, update_command_permissions
from .slash.types import AdditionalType, BaseCommand, CommandType, ContextCommand, MessageCommand, OptionType, SlashCommand, SlashOption, SlashSubcommand, UserCommand
from .tools import MISSING, _none, _or, get_index, setup_logger, get
from .http import jsonifyMessage, BetterRoute, send_files
from .receive import ComponentContext, Interaction, Message, PressedButton, SelectedMenu, SlashedContext, WebhookMessage, SlashedCommand, SlashedSubCommand, getMessage
from .override import override_dpy as override_it
from .listener import Listener

import discord
from discord.ext import commands as com
from discord.ext.commands.errors import CommandNotFound
from discord.errors import Forbidden, InvalidArgument, NotFound

import zlib
import json
import inspect
import asyncio
import contextlib
from typing import Coroutine, Dict, List, Tuple, Union
try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

logging = setup_logger(__name__)

class Slash():
    """
    A class for using slash commands
    
    Parameters
    ----------
        client: :class:`commands.Bot`
            The bot client

        parse_method: :class:`bool`, optional
            How received option data should be treated; Default ``ParseMethod.AUTO``

        delete_unused: :class:`bool`, optional
            Whether the commands that are not registered by this slash extension should be deleted in the api; Default ``False``

        sync_on_cog: :class:`bool`, optional
            Whether the slashcommands should be updated whenever a new cog is added or removed; Default ``False``

        wait_sync: :class:`float`, optional
            How many seconds will be waited until the commands are going to be synchronized; Default ``1``

        auto_defer: Tuple[:class:`bool`, :class:`bool`]
            Settings for the auto-defer; Default ``(True, False)``

            ``[0]``: Whether interactions should be deferred automatically

            ``[1]``: Whether the deferration should be hidden (True) or public (False)

    Example
    ------------------
    Example for using slash commands

    .. code-block::

        ...
        # Your bot declaration and everything
        slash = Slash(client)

    For creating a slash command use

    .. code-block::

        ...
        @slash.command(name="my_command", description="this is my slash command", options=[SlashOption(str, "option", "this is an option")])
        async def command(ctx: SlashedCommand):
            ...
    
    And for subcommand groups use

    .. code-block::

        ...
        @slash.subcommand_group(base_names=["base", "group"], name="sub", description="this is a sub command group")
        async def subgroup(ctx: SlashedSubCommand):
            ...
        

    """
    def __init__(self, client, parse_method = ParseMethod.AUTO, auto_sync=True, delete_unused = False, sync_on_cog=False, wait_sync = 1, auto_defer = False) -> None:
        """
        Creates a new slash command thing
        
        Example
        ```py
        Slash(client)
        ```
        """
        # disable decompressing
        # self._buffer = bytearray()
        # self._zlib = zlib.decompressobj()

        self.ready = False
        self.parse_method: int = parse_method
        self.delete_unused: bool = delete_unused
        self.wait_sync: float = wait_sync
        self.auto_defer: Tuple[bool, bool] = (auto_defer, False) if isinstance(auto_defer, bool) else auto_defer
        self.auto_sync = auto_sync

        self._discord: com.Bot = client
        self.commands: Dict[(str, SlashCommand)] = {}
        self.subcommands: Dict[(str, Dict[(str, Union[dict, SlashSubcommand])])] = {}
        self.context_commands: Dict[str, ContextCommand] = {"message": {}, "user": {}}
        if discord.__version__.startswith("2"):
            self._discord.add_listener(self._on_slash_response, "on_socket_raw_receive")
        elif discord.__version__.startswith("1"):
            self._discord.add_listener(self._on_slash_response, 'on_socket_response')

        
        old_add = self._discord.add_cog
        def add_cog_override(*args, **kwargs):
            cog = args[0] if len(args) > 0 else kwargs.get("cog")
            for com in self._get_cog_commands(cog):
                com.cog = cog
                self._add_to_cache(com)
            old_add(*args, **kwargs)
            if self.ready is True and sync_on_cog is True:
                self._discord.loop.create_task(self.sync_commands(self.delete_unused))
        self._discord.add_cog = add_cog_override

        old_remove = self._discord.remove_cog
        def remove_cog_override(*args, **kwargs):
            cog = args[0] if len(args) > 0 else kwargs.get("cog")
            for com in self._get_cog_commands(cog):
                com.cog = cog
                self._remove_from_cache(com)
            old_remove(*args, **kwargs)
            if self.ready is True and sync_on_cog is True:
                self._discord.loop.create_task(self.sync_commands(self.delete_unused))
        self._discord.remove_cog = remove_cog_override
        
        async def on_connect():
            self.ready = True
            if self.auto_sync is False:
                return
            await asyncio.sleep(_or(self.wait_sync, 1))
            await self.sync_commands(self.delete_unused)
        self._discord.add_listener(on_connect)

    async def _on_slash_response(self, msg):
        if discord.__version__.startswith("2"):
            # comment decompressing out in canse it is needed in the future for some reason
            if isinstance(msg, bytes):
                raise NotImplementedError("decompressing was removed! Please upgrade your discord.py version")
            #     try:
            #         self._buffer.extend(msg)
            #         if len(msg) < 4 or msg[-4:] != b'\x00\x00\xff\xff':
            #             return
            #         msg = self._zlib.decompress(self._buffer)
            #         msg = msg.decode('utf-8')
            #         self._buffer = bytearray()
            #     except:
            #         return
            if isinstance(msg, str):
                msg = json.loads(msg)

        if msg["t"] != "INTERACTION_CREATE":
            return
        data = msg["d"]

        if int(data["type"]) not in [1, 2]:
            return

        user = discord.Member(data=data["member"], guild=self._discord._connection._get_guild(int(data["guild_id"])), state=self._discord._connection) if data.get("member") is not None else discord.User(state=self._discord._connection, data=data["user"])
        if user.dm_channel is None:
            await user.create_dm()

        interaction = Interaction(self._discord._connection, data, user)
        if self.auto_defer[0] is True:
            await interaction.defer(self.auto_defer[1])
        self._discord.dispatch("interaction_received", interaction)


        #region basic commands
        if CommandType(data["data"]["type"]) is CommandType.Slash and not (data["data"].get("options") and data["data"]["options"][0]["type"] in [OptionType.SUB_COMMAND, OptionType.SUB_COMMAND_GROUP]):
            x = self.commands.get(data["data"]["name"])
            if x is None:
                cog_commands = {}
                for x in self._discord.cogs:
                    cog_commands = self._get_cog_slash_commands(x)
                x = cog_commands.get(data["data"]["name"])
            if x is not None:
                options = {}
                if data["data"].get("options") is not None:
                    options = await handle_options(data, data["data"]["options"], self.parse_method, self._discord)
                context = SlashedCommand(self._discord, command=x, data=data, user=user, args=options)
                context._patch(x)
                # Handle autodefer
                context._handle_auto_defer(self.auto_defer)
                self._discord.dispatch("slash_command", context)
                if hasattr(x, "invoke"):
                    await x.invoke(context, **options)
                elif x.callback is not None:
                    await x.callback(context, **options)
                return
        elif CommandType(data["data"]["type"]) is CommandType.User:
            x = self.context_commands["user"].get(data["data"]["name"])
            if x is not None:
                member = await handle_thing(data["data"]["target_id"], OptionType.MEMBER, data, self.parse_method, self._discord)
                context = SlashedContext(self._discord, command=x, data=data, user=user, param=member)
                # Handle autodefer
                context._handle_auto_defer(self.auto_defer)
                context._patch(x)

                self._discord.dispatch("context_command", context, member)
                if x.callback is not None:
                    if hasattr(x, "invoke"):
                        await x.invoke(context, member)
                    else:
                        await x.callback(context, member)
                return
        elif CommandType(data["data"]["type"]) is CommandType.Message:
            x = self.context_commands["message"].get(data["data"]["name"])
            if x is not None:
                message = await handle_thing(data["data"]["target_id"], 44, data, self.parse_method, self._discord)
                context = SlashedContext(self._discord, command=x, data=data, user=user, param=message)
                context._patch(x)
                # Handle autodefer
                context._handle_auto_defer(self.auto_defer)
                
                self._discord.dispatch("context_command", context, message)
                if x.callback is not None:
                    if hasattr(x, "invoke"):
                        await x.invoke(context, message)
                    else:
                        await x.callback(context, message)
                return
        #endregion

        fixed_options = []
        x_base = self.subcommands.get(data["data"]["name"])
        if not x_base:
            cog_commands = {}
            for x in self._discord.cogs:
                cog_commands = self._get_cog_subgroup_commands(x)
            x_base = cog_commands.get(data["data"]["name"])

        if x_base:
            op = data["data"]["options"][0]
            while op["type"] != 1:
                op = op["options"][0]
            fixed_options = op.get("options", [])
            
            x = x_base.get(data["data"]["options"][0]["name"])
            if isinstance(x, dict):
                x = x.get(data["data"]["options"][0]["options"][0]["name"])

            options = await handle_options(data, fixed_options, self.parse_method, self._discord)

            if x:
                context = SlashedSubCommand(self._discord, x, data, user, options)
                context._patch(x)
                # Handle auto_defer
                context._handle_auto_defer(self.auto_defer)

                self._discord.dispatch("slash_command", context)
                if hasattr(x, "invoke"):
                    await x.invoke(context, **options)
                elif x.callback is not None:
                    await x.callback(context, **options)
                return
    

    def _get_cog_commands(self, cog):
        # Get all BaseCallables with __type__ of 1 (Slash)
        return [x[1] for x in inspect.getmembers(cog, lambda x: isinstance(x, BaseCallable) and x.__type__ == 1)]
        
    def _get_commands_from_cog(self, cog, cls):
        commands = {x.name: x for x in self._get_cog_commands(cog) if isinstance(x, cls)}
        for x in commands:
            commands[x].cog = cog
        if cls is CogSubCommandGroup:
            back = commands
            commands = {}
            for _, x in list(back.items()):
                commands[x.base_names[0]] = {}
                if len(x.base_names) > 1:
                    commands[x.base_names[0]][x.base_names[1]] = {x.name: x}
                else:
                    commands[x.base_names[0]] = {x.name: x}
        return commands
    def _get_cog_slash_commands(self, name):
        """Returns all slash commands registered in the cogs"""
        return self._get_commands_from_cog(self._discord.get_cog(name), CogCommand)
    def _get_cog_subgroup_commands(self, name):
        return self._get_commands_from_cog(self._discord.get_cog(name), CogSubCommandGroup)


    async def sync_commands(self, delete_unused=None):
        """
        Synchronizes the slash commands with the api
        
        Parameters
        ----------
            delete_unused: :class:`bool`, optional
                Whether the unused command should be deleted from the api; default ``None``
        
        Raises
        ------
            :raises: :class:`InvalidArgument` : A slash command has an invalid guild_id
            :raises: :class:`InvalidArgument` : A slash command has an invalid id specified in the guild_permissions
        
        """
        delete_unused = delete_unused or self.delete_unused
        added_commands = {
            "globals": {},
            "guilds": {}
        }
        own_guild_ids = [x.id for x in self._discord.guilds]
        
        commands = self.gather_commands()
        async def guild_stuff(command, guild_ids, is_change=False):
            """Adds the command to the guilds"""
            for x in guild_ids:
                if str(x) in list(command.__guild_changes__.keys()) and is_change is False:
                    continue
                if command.guild_permissions is not None:
                    for x in list(command.guild_permissions.keys()):
                        if int(x) not in own_guild_ids:
                            raise InvalidArgument("guild_permissions invalid! Client is not in a guild with the id " + str(x))
                if int(x) not in own_guild_ids:
                    raise InvalidArgument("guild_ids invalid! Client is not in a server with the id '" + str(x) + "'")
            
                if added_commands["guilds"].get(x) is None:
                    added_commands["guilds"][x] = {}

                if command.guild_permissions is not None:
                    command.permissions = command.guild_permissions.get(x)
                
                await self.add_guild_command(command, x)
                added_commands["guilds"][x][command.name] = command
        async def global_stuff(command):
            await self.add_global_command(command)
            added_commands["globals"][command.name] = command

        async def add_command(command):
            # If command is set to no sync
            if command.__sync__ == False:
                return
            if command.__guild_changes__ != {}:
                for guild_id in command.__guild_changes__:
                    _name, _description, _default_permission = command.__guild_changes__[guild_id]
                    cur = command.copy()
                    if _name is not None:
                        cur.name = _name
                    if _description is not None:
                        cur.description = _description
                    if _default_permission is not None:
                        cur.default_permission = _default_permission
                    cur.guild_ids = [int(guild_id)]
                    await guild_stuff(cur, [int(guild_id)], is_change=True)
            # guild only command
            if command.guild_ids is not None:
                logging.debug("adding '" + str(command.name) + "' as guild_command")
                await guild_stuff(command, command.guild_ids)
            # global command with guild permissions
            elif command.guild_ids is None and command.guild_permissions is not None:
                logging.debug("adding '" + str(command.name) + "' as guild_command and global")
                await global_stuff(command)
                for guild in list(command.guild_permissions.keys()):
                    await self.update_permissions(command.name, command.type, guild_id=guild, permissions=command.guild_permissions[guild], global_command=True)
            # global command
            elif command.guild_ids is None and command.guild_permissions is None:
                logging.debug("adding '" + str(command.name) + "' as global command")
                await global_stuff(command)

        for x in commands:
            await add_command(commands[x])

        for x in self.context_commands:
            for com in self.context_commands[x]:
                await add_command(self.context_commands[x][com])

        if delete_unused:
            api_coms = await self._get_global_commands()
            for apic in api_coms:
                logging.debug("deleting global command '" + str(apic["name"]) + "'")
                if added_commands["globals"].get(apic["name"]) is None:
                    await delete_global_command(self._discord, apic["id"])
            async for x in self._discord.fetch_guilds():
                _id = str(x.id)
                api_coms = await self._get_guild_commands(_id)
                for apic in api_coms:
                    if added_commands["guilds"].get(int(_id)) is not None:
                        _id = int(_id)
                    elif added_commands["guilds"].get(str(_id)) is not None:
                        _id = str(_id)
                    if added_commands["guilds"].get(_id) is None or added_commands["guilds"][_id].get(apic["name"]) is None:
                        logging.debug("deleting guild command '" + str(apic["name"]) + "' in guild " + str(_id))
                        await delete_guild_command(self._discord, apic["id"], _id)

        logging.info("synchronized slash commands")
    
    async def _get_api_command(self, name) -> Union[dict, None]:
        for x in await self._get_commands():
            if x["name"] == name:
                return x
    async def _get_guild_api_command(self, name, typ, guild_id) -> Union[dict, None]:
        for x in await self._get_guild_commands(guild_id):
            if hasattr(typ, "value"):
                typ = typ.value
            if x["name"] == name and x["type"] == typ:
                return x
    async def _get_global_api_command(self, name, typ) -> Union[dict, None]:
        for x in await self._get_global_commands():
            if x["name"] == name and x["type"] == typ:
                return x

    async def _get_commands(self) -> List[dict]:
        return await self._get_global_commands() + await self._get_all_guild_commands()
    async def _get_global_commands(self) -> List[dict]:
        return await get_global_commands(self._discord)
    async def _get_all_guild_commands(self):
        commands = []
        async for x in self._discord.fetch_guilds():
            try:
                commands += await get_guild_commands(self._discord, x.id)
            except Forbidden:
                logging.warn("Got forbidden in " + str(x.name) + " (" + str(x.id) + ")")
                continue
        return commands
    async def _get_guild_commands(self, guild_id: str) -> List[dict]:
        logging.debug("getting guild commands in " + str(guild_id))
        return await get_guild_commands(self._discord, guild_id)
    
    def gather_commands(self) -> Dict[str, SlashCommand]:
        commands = self.commands.copy()
        for _base in self.subcommands:
            # get first base
            for _sub in self.subcommands[_base]:
                # get second base/command
                sub = self.subcommands[_base][_sub]
                # when command has subcommand groups
                if isinstance(sub, dict):
                    for _group in self.subcommands[_base][_sub]:
                        # the subcommand group
                        group = self.subcommands[_base][_sub][_group]
                        # if there's already a base command
                        if commands.get(_base) is not None:
                            # Check if base already has an option with the subs name
                            index = get_index(commands[_base].options, _sub, lambda x: getattr(x, "name"))
                            # if first base_name already exists
                            if index > -1:
                                # add to sub options
                                base_ops = commands[_base].options
                                base_ops[index].options += [group.to_option()]
                                commands[_base].options = base_ops
                            # if not exists
                            else:
                                # create sub option + group option
                                commands[_base].options += [SlashOption(OptionType.SUB_COMMAND_GROUP, _sub, options=[group.to_option()])]
                        # if no base command
                        else:
                            # create base0 command together with base1 option and groupcommand option
                            commands[_base] = SlashCommand(None, _base, None, [
                                    SlashOption(OptionType.SUB_COMMAND_GROUP, _sub, options=[group.to_option()])
                                ],
                                guild_ids=group.guild_ids, default_permission=group.default_permission, guild_permissions=group.guild_permissions)
                # if is basic subcommand
                else:
                    # If base exists
                    if commands.get(_base) is not None:
                        commands[_base].options += [sub.to_option()]
                    else:
                        # create base0 command with name option
                        commands[_base] = SlashCommand(None, _base, options=[sub.to_dict()], guild_ids=sub.guild_ids, default_permission=sub.default_permission, guild_permissions=sub.guild_permissions)
        return commands

    async def create_command(self, command) -> SlashCommand:
        """
        Adds a command to the api. You shouldn't use this method unless you know what you're doing
        
        Parameters
        ----------
            command: :class:`SlashCommand` | :class:`ContextCommand`
                The command that should be added
        
        Raises
        ------
            :raises: :class:`InvalidArgument` : When a guild-id in ``guild_ids`` is not a valid server where the bot client is in it
            :raises: :class:`InvalidArgument` : When a guild-id in ``guild_permissions`` is not a valid server where the bot client is in it
        
        """
        if command.guild_ids is not None:
            guild_ids = command.guild_ids
            own_guild_ids = [x.id for x in self._discord.guilds]
            for x in guild_ids:
                if command.guild_permissions is not None:
                    for x in list(command.guild_permissions.keys()):
                        if int(x) not in own_guild_ids:
                            raise InvalidArgument("guild_permissions invalid! Client is not in a guild with the id " + str(x))
                if int(x) not in own_guild_ids:
                    raise InvalidArgument("guild_ids invalid! Client is not in a server with the id '" + str(x) + "'")

                if command.guild_permissions is not None:
                    command.permissions = command.guild_permissions.get(x)
                
                await self.add_guild_command(command, x)
        else:
            await self.add_global_command(command)
        return command
    async def add_global_command(self, base):
        """
        Adds a slash command to the global bot commands
        
        Parameters
        ----------
            base: :class:`~SlashCommand`
                The slash command to add
        
        """
        api_command = await self._get_global_api_command(base.name, base._json["type"])
        if api_command is None:
            await create_global_command(base.to_dict(), self._discord)
        else:
            if api_command != base:
                await edit_global_command(api_command["id"], self._discord, base.to_dict())
    async def add_guild_command(self, base, guild_id):
        """
        Adds a slash command to a guild
        
        Parameters
        ----------
            base: :class:`~SlashCommand`
                The guild slash command which should be added
            guild_id: :class:`str`
                The ID of the guild where the command is going to be added
        
        """
        target_guild = guild_id
        api_command = await self._get_guild_api_command(base.name, base._json["type"], guild_id)
        if api_command is not None:
            api_permissions = await get_command_permissions(self._discord, api_command["id"], guild_id)
        global_command = await self._get_global_api_command(base.name, base._json["type"])
        # If no command in that guild or a global one was found
        if api_command is None or global_command is not None:
            # # Check global commands
            # If global command exists, it will be deleted
            if global_command is not None:
                await delete_global_command(self._discord, global_command["id"])
            await create_guild_command(base.to_dict(), self._discord, target_guild, base.permissions.to_dict())
        elif api_command != base:
            await edit_guild_command(api_command["id"], self._discord, target_guild, base.to_dict(), base.permissions.to_dict())
        elif api_permissions != base.permissions:
            await update_command_permissions(self._discord.user.id, self._discord.http.token, guild_id, api_command["id"], base.permissions.to_dict())

    async def update_permissions(self, name, typ: Literal["slash", 1, "user", 2, "message", 3] = 1, *, guild_id=None, default_permission=None,  permissions=None, global_command=False):
        """
        Updates the permissions for a command
        
        Parameters
        ----------
            name: :class:`str`
                The name of the command that should be updated
            typ: :class:`str` | :class:`int`
                The type of the command (one of ``"slash", CommandType.Slash, "user", CommandType.User, "message", CommandType.Message``)
            default_permission: :class:`bool`, optional
                The default permission for the command if users can use it
            guild_id: :class:`int` | :class:`str`, optional
                The ID of the guild where the permissions should be updated.
                    This needs to be passed when you use the `permissions` parameter or want to update a guild command
            permissions: :class:`SlashPermission`, optional
                The new permissions for the command
            global_command: :class:`bool`, optional
                If the command is a global command or a guild command; default ``False``
        
        """
        if guild_id is not None:
            guild_id = int(guild_id)
        typ = CommandType.from_string(typ).value
        if global_command is True:
            api_command = await self._get_global_api_command(name, typ)
        else:
            api_command = await self._get_guild_api_command(name, typ, guild_id)
        if api_command is None:
            raise CommandNotFound("Slash command with name " + str(name) + " and type " + str(typ) + " not found in the api!")
        if permissions is not None:
            await update_command_permissions(self._discord.user.id, self._discord._connection.http.token, guild_id, api_command["id"], permissions.to_dict())
        if default_permission is not None:
            default_permission = default_permission or False
            api_command["default_permission"] = default_permission
            if global_command is True:
                await edit_global_command(api_command["id"], self._discord, api_command)
            else:
                await edit_guild_command(api_command["id"], self._discord, guild_id, api_command)
    async def edit_command(self, old_name, typ: Literal["slash", 1, "user", 2, "message", 3] = 1, guild_id=None, *, name, description, options, guild_ids, default_permission, guild_permissions, callback=MISSING):
        """
        Edits a command
        
        Parameters
        ----------
            old_name: :class:`str`
                The original name of the command
            typ: Literal[``'slash'``, ``1`` | ``'user'``, ``2`` | ``'message'``, ``3``], optional
                The type of the command; default ``1`` (slashcommand)
            guild_id: :class:`[type]`, optional
                The guild id of the command; default ``MISSING``
            name: :class:`str`, optional
                The new name of the command
            description: :class:`str`, optional
                The new description of the command
            options: List[:class:`~SlashOption`], optional
                The new
            guild_ids: List[:class:`int` | :class:`str`], optional
                The list of new guild_ids where the command is available
            default_permission: :class:`bool`, optional
                The new default permissions for the command
            guild_permissions: :class:`dict`, optional
                The new guild permissions for the command
            callback: :class:`function`, optional, optional
                The new command callback; default ``MISSING``
        
        Raises
        ------
            :raises: :class:`NotFound` : When a command in the internal cache doesn't exsist
            :raises: :class:`NotFound` : When a command in the api doesn't exist
        
        """
        typ = CommandType.from_string(typ).value
        old_name = format_name(old_name)

        old_command = self._get_command(old_name, typ)
        if old_command is None:
            raise NotFound("Could not find a command in the internal cache with the name " + str(old_name) + " and the type " + str(typ))
        command = old_command
        if guild_id is not None:
            api_command = await self._get_guild_api_command(old_name, typ, guild_id)
        else:
            api_command = await self._get_global_api_command(old_name, typ)
        if old_command is None:
            raise NotFound("Could not find a command in the api with the name " + str(old_name) + " and the type " + str(typ))

        if name is not None:
            command.name = name
        if description is not None:
            command.description = _or(description, inspect.getdoc(callback).split("\n")[0] if callback is not None and inspect.getdoc(callback) is not None else None, name, old_name)
        if options is not None:
            command.options = options
        if default_permission is not None:
            command.default_permission = default_permission
        if guild_permissions is not None:
            command.guild_permissions = guild_permissions
        if guild_ids is not None:
            command.guild_ids = guild_ids
        if callback is not MISSING:
            command.callback = callback
        # When command only should be edited in one guild
        if guild_id is not None and guild_ids is None or old_command.guild_ids == guild_ids:
            await edit_guild_command(api_command["id"], self._discord, guild_id, command.to_dict(), command.guild_permissions.get(guild_id))
        # When guild_ids were changed
        elif guild_ids is not None and guild_ids != old_command.guild_ids:
            for x in guild_ids:
                await self.add_guild_command(command, x)
        # When guild command is changed to global command
        elif guild_id is not None and guild_ids is None and old_command.guild_ids is not None:
            for x in old_command.guild_ids:
                com = await self._get_guild_api_command(old_command.name, old_command.command_type, x)
                await delete_guild_command(self._discord, com["id"], x)
            await self.add_global_command(command)
        # When global command is changed to guild command
        elif old_command.guild_ids is None and command.guild_ids is not None:
            com = await self._get_global_api_command(old_command.name, old_command.command_type)
            await delete_global_command(self._discord, com["id"])
            await self.create_command(command)
        else:
            await self.create_command(command)
        self._set_command(old_name, command)
    async def edit_subcommand(self, base_names, old_name, guild_id=MISSING, *, name=None, description=None, options=None, guild_ids=MISSING, default_permission=None, guild_permissions=None, callback=MISSING):
        """
        Edits a subcommand
        
        Parameters
        ----------
            base_names: List[:class:`str`] | :class:`str`
                The base_names of the command
            old_name: :class:`str`
                The original name of the command
            guild_id: :class:`[type]`, optional
                The guild id of the command where the changes should be applied; default ``MISSING``
            name: :class:`str`, optional
                The new name of the command
            description: :class:`str`, optional
                The new description of the command
            options: List[:class:`~SlashOption`], optional
                The new options for the command
            guild_ids: List[:class:`int` | :class:`str`], optional
                The list of new guild_ids where the command is available
            default_permission: :class:`bool`, optional
                The new default permissions for the command
            guild_permissions: :class:`dict`, optional
                The new guild permissions for the command
            callback: :class:`function`, optional, optional
                The new command callback; default ``MISSING``
        
        Raises
        ------
            :raises: :class:`NotFound` : When a command in the internal cache doesn't exsist
            :raises: :class:`NotFound` : When a command in the api doesn't exist
        
        """
        if isinstance(base_names, str):
            base_names = [base_names]
        base = self.gather_commands()[base_names[0]]
        origin_base = base
        if len(base_names) > 1 and isinstance(base, dict):
            base = get(base.options, base_names[1], lambda x: x.getattr("name"))
        op: SlashOption = get(base, old_name, lambda x: x.getattr("name"))
        sub = SlashSubcommand.from_data(op.to_dict())

        if guild_id is not None:
            api_command = await self._get_guild_api_command(old_name, CommandType.Slash, guild_id)
        else:
            api_command = await self._get_global_api_command(old_name, CommandType.Slash)
        
        if name is not None:
            sub.name = name
        if description is not None:
            sub.description = description
        if options is not None:
            sub.options = options
        if guild_ids is not None:
            sub.guild_ids = guild_ids
        if default_permission is not None:
            sub.default_permission = default_permission
        if guild_permissions is not None:
            sub.guild_permissions = guild_permissions
        if callback is not MISSING:
            sub.callback = callback

        base.guild_ids = sub.guild_ids
        base.guild_permissions = sub.guild_permissions
        base.default_permission = default_permission

        # When command only should be edited in one guild
        if guild_id is not MISSING and guild_ids is None or sub.guild_ids == guild_ids:
            await edit_guild_command(api_command["id"], self._discord, guild_id, base.to_dict(), sub.guild_permissions.get(guild_id))
        # When guild_ids were changed
        elif guild_ids is not MISSING and guild_ids != origin_base.guild_ids:
            for x in guild_ids:
                await self.add_guild_command(base, x)
        # When guild command is changed to global command
        elif guild_id is not MISSING and guild_ids is None and base.guild_ids is not MISSING:
            for x in base.guild_ids:
                com = await self._get_guild_api_command(base.name, base.command_type, x)
                await delete_guild_command(self._discord, com["id"], x)
            await self.add_global_command(base)
        # When global command is changed to guild command
        elif origin_base.guild_ids is None and base.guild_ids is not MISSING:
            com = await self._get_global_api_command(origin_base.name, origin_base.command_type)
            await delete_global_command(self._discord, com["id"])
            await self.create_command(base)
        else:
            await self.create_command(base)
        self._set_command(old_name, sub)

    def _get_command(self, name, typ: Literal["slash", 1, "user", 2, "message", 3]) -> SlashCommand:
        typ = CommandType.from_string(typ)
        if typ is CommandType.Slash:
            return self.commands.get(name)
        elif typ is CommandType.User:
            return self.context_commands["user"].get(name)
        else:
            return self.context_commands["message"].get(name)
    def _set_command(self, old_name, command: SlashCommand):
        if command.command_type is CommandType.Slash:
            if self.commands.get(old_name) is not None:
                del self.commands[old_name]
            self.commands[command.name] = command
        elif command.command_type is CommandType.Message:
            del self.context_commands["message"][old_name]
            self.context_commands["message"][command.name] = command
        else:
            del self.context_commands["user"][old_name]
            self.context_commands["user"][command.name] = command
    def _add_to_cache(self, base: Union[SlashCommand, SlashSubcommand], is_base=False):
        if base.has_aliases and is_base is False:
            for a in base.__aliases__:
                cur = base.copy()
                cur.name = a
                self._add_to_cache(cur, is_base=True)
        if base.__guild_changes__ != {} and is_base is False:
            for guild_id in base.__guild_changes__:
                _name, _description, _default_permission = base.__guild_changes__.get(guild_id)
                cur = base.copy()
                if _name is not None:
                    cur.name = _name
                if _description is not None:
                    cur.description = _description
                if _default_permission is not None:
                    cur.default_permission = _default_permission
                cur.guild_ids = [int(guild_id)]
                self.commands[cur.name] = cur
        if base.command_type is CommandType.Slash:
            # basic slash command
            if isinstance(base, SlashCommand):
                self.commands[base.name] = base
            # subcommand or subgroup
            else:
                # when subcommands is missing base
                if self.subcommands.get(base.base_names[0]) is None:
                    self.subcommands[base.base_names[0]] = {}
                # subgroup
                if len(base.base_names) > 1:
                    # if cache is missing second base_name 
                    if self.subcommands[base.base_names[0]].get(base.base_names[1]) is None:
                        self.subcommands[base.base_names[0]][base.base_names[1]] = {}
                    # add to internal cache
                    self.subcommands[base.base_names[0]][base.base_names[1]][base.name] = base
                # subcommand
                else:
                    # add to cache
                    self.subcommands[base.base_names[0]][base.name] = base
        # context user command
        elif base.command_type is CommandType.User:
            self.context_commands["user"][base.name] = base
        # context message command
        elif base.command_type is CommandType.Message:
            self.context_commands["message"][base.name] = base
    def _remove_from_cache(self, base: Union[SlashCommand, SlashSubcommand]):
        # slash command
        if base.command_type is CommandType.Slash:
            # basic slash command
            if isinstance(base, (SlashCommand, CogCommand)):
                del self.commands[base.name]
            # subcommand or subgroup
            else:
                # subgroup
                if len(base.base_names) > 1:
                    # remove from internal cache
                    del self.subcommands[base.base_names[0]][base.base_names[1]][base.name]
                # subcommand
                else:
                    # remove from cache
                    del self.subcommands[base.base_names[0]][base.name]
        # context user command
        elif base.command_type is CommandType.User:
            # remove from cache
            del self.context_commands["user"][base.name]
        # context message command
        elif base.command_type is CommandType.Message:
            # remove from cache
            del self.context_commands["message"][base.name]

    async def delete_global_commands(self):
        """**Deletes all global commands**"""
        await delete_global_commands(self._discord)
    async def delete_guild_commands(self, guild_id: str):
        """
        **Deletes all commands in a guild**

        Parameters
        ----------
            guild_id: :class:`str`
                The id of the guild where all commands are going to be deleted
        
        """
        await delete_guild_commands(self._discord, guild_id)
    async def nuke_commands(self):
        """**Deletes every command for the bot, including globals and commands in every guild**"""
        logging.debug("nuking...")
        await self.delete_global_commands()
        logging.debug("nuked global commands")
        async for guild in self._discord.fetch_guilds():
            logging.debug("nuking commands in" + str(guild.id))
            await self.delete_guild_commands(guild.id)
            logging.debug("nuked commands in" + str(guild.name) + " (" + str(guild.id) + ")")
        logging.info("nuked all commands")


    def add_command(self, name=None, callback=None, description=None, options=None, guild_ids=None, default_permission=True, guild_permissions=None, api=False) -> Union[None, Coroutine]:
        """
        Adds a new slashcommand

        name: :class:`str`
            1-32 characters long name; default MISSING

            .. note::

                The name will be corrected automaticaly (spaces will be replaced with "-" and the name will be lowercased)
        callback: :class:`function`, optional
            A callback that will be called when the command was received
        description: :class:`str`, optional
            1-100 character description of the command; default the command name
        options: List[:class:`~SlashOptions`], optional
            The parameters for the command; default MISSING
        choices: List[:class:`tuple`] | List[:class:`dict`], optional
            Choices for string and int types for the user to pick from; default MISSING
        guild_ids: List[:class:`str` | :class:`int`], optional
            A list of guild ids where the command is available; default MISSING
        default_permission: :class:`bool`, optional
            Whether the command can be used by everyone or not
        guild_permissions: Dict[``guild_id``: :class:`~SlashPermission`]
            The permissions for the command in guilds
                Format: ``{"guild_id": SlashPermission}``
        api: :class:`bool`, optional
            Whether the command should be registered to the api (True) or just added in the internal cache
                If it's added to the internal cache, it will be registered to the api when calling the `sync_commands` function.
                If ``api`` is True, this function will return a promise
        """
        command = SlashCommand(callback, name, description, options, guild_ids=guild_ids, default_permission=default_permission, guild_permissions=guild_permissions)
        self._add_to_cache(command)
        if api is True:
            if self.ready is False:
                raise Exception("Slashcommands are not ready yet")
            return self.create_command(command) 
        return command
    def command(self, name=None, description=None, options=None, guild_ids=None, default_permission=True, guild_permissions=None):
        """
        A decorator for a slash command
        
        command in discord:
            ``/name [options]``

        Parameters
        ----------
            name: :class:`str`, optional
                1-32 characters long name; default MISSING
                
                .. note::

                    The name will be corrected automaticaly (spaces will be replaced with "-" and the name will be lowercased)
            
            description: :class:`str`, optional
                1-100 character description of the command; default the command name
            options: List[:class:`~SlashOptions`], optional
                The parameters for the command; default MISSING
            choices: List[:class:`tuple`] | List[:class:`dict`], optional
                Choices for string and int types for the user to pick from; default MISSING
            guild_ids: List[:class:`str` | :class:`int`], optional
                A list of guild ids where the command is available; default MISSING
            default_permission: :class:`bool`, optional
                Whether the command can be used by everyone or not
            guild_permissions: Dict[``guild_id``: :class:`~SlashPermission`]
                The permissions for the command in guilds
                    Format: ``{"guild_id": SlashPermission}``

        Decorator
        ---------

            callback: :class:`method(ctx)`
                The asynchron function that will be called if the command was used
                    ctx: :class:`~SlashedCommand`
                        The used slash command

                    .. note::

                        ``ctx`` is just an example name, you can use whatever you want for that

        Example
        -------
        .. code-block::

            @slash.command(name="hello_world", description="This is a test command", 
            options=[
                SlashOption(str, name="parameter", description="this is a parameter", choices=[("choice 1", "test")])
            ], guild_ids=[785567635802816595], default_permission=False, 
            guild_permissions={
                    785567635802816595: SlashPermission(allowed={"539459006847254542": SlashPermission.USER})
                }
            )
            async def command(ctx, parameter = None):
                ...
        """
        def wrapper(callback):
            """
            The wrapper for the callback function. The function's parameters have to have the same name as the parameters specified in the slash command.

            `ctx` is of type :class:`~SlashedCommand` and is used for responding to the interaction and more

            Examples
            --------
            - no parameter:
                `async def command(ctx): ...`
            - required parameter "number":
                `async def command(ctx, number): ...`
            - optional parameter "user":
                `async def command(ctx, user=default_value)`
            - multiple optional parameters "user", "number":
                `async def command(ctx, user=default_value, number=default_value)`
            - one required and one optional parameter "user", "text":
                `async def command(ctx, user, text=default_value)`

            Note: Replace `default_value` with a value you want to be used if the parameter is not specified in discord, if you don't want a default value, just set it to `None`
            """
            return self.add_command(name, callback, description, options, guild_ids, default_permission, guild_permissions)
        return wrapper
    def add_subcommand(self, base_names, name=None, callback=None, description=None, options=None, guild_ids=None, default_permission=True, guild_permissions=None):
        command = SlashSubcommand(callback, base_names, name, description, options, guild_ids=guild_ids, default_permission=default_permission, guild_permissions=guild_permissions)
        self._add_to_cache(command)
    def subcommand(self, base_names, name=None, description=None, options=[], guild_ids=None, default_permission=True, guild_permissions=None):
        """
        A decorator for a subcommand group
        
        command in discord
            ``/base_names... name [options]``

        Parameters
        ----------
            base_names: List[:class:`str`] | :class:`str`
                The names of the parent bases, currently limited to 2
                    If you want to make a subcommand (``/base name``), you have to use a str instead of a list
            name: :class:`str`, optional
                1-32 characters long name; default MISSING
                
                .. note::

                    The name will be corrected automaticaly (spaces will be replaced with "-" and the name will be lowercased)
            description: :class:`str`, optional
                1-100 character description of the command; default the command name
            options: List[:class:`~SlashOptions`], optional
                The parameters for the command; default MISSING
            choices: List[:class:`tuple`] | List[:class:`dict`], optional
                Choices for string and int types for the user to pick from; default MISSING
            guild_ids: List[:class:`str` | :class:`int`], optional
                A list of guild ids where the command is available; default MISSING
            default_permission: :class:`bool`, optional
                Whether the command can be used by everyone or not
            guild_permissions: Dict[``guild_id``: :class:`~SlashPermission`]
                The permissions for the command in guilds
                    Format: ``{"guild_id": SlashPermission}``

            .. note::

                Permissions will be the same for every subcommand with the same base

        Decorator
        ---------

            callback: :class:`method(ctx)`
                The asynchron function that will be called if the command was used
                    ctx: :class:`~SlashedSubCommand`
                        The used slash command

                    .. note::

                        ``ctx`` is just an example name, you can use whatever you want for that
        
        Example
        -------
        
        subcommand

        .. code-block::

            @slash.subcommand_group(base_names="hello", name="world", options=[
                SlashOption(argument_type="user", name="user", description="the user to tell the holy words")
            ], guild_ids=[785567635802816595])
            async def command(ctx, user):
                ...

        subcommand-group

        .. code-block::

            @slash.subcommand_group(base_names=["hello", "beautiful"], name="world", options=[
                SlashOption(argument_type="user", name="user", description="the user to tell the holy words")
            ], guild_ids=[785567635802816595])
            async def command(ctx, user):
                ...

        """
        def wrapper(callback):
            """
            The wrapper for the callback function. The function's parameters have to have the same name as the parameters specified in the slash command.

            `ctx` is of type :class:`~SlashedCommand` and is used for responding to the interaction and more

            Examples
            --------
            - no parameter:
                `async def command(ctx): ...`
            - required parameter "number":
                `async def command(ctx, number): ...`
            - optional parameter "user":
                `async def command(ctx, user=default_value)`
            - multiple optional parameters "user", "number":
                `async def command(ctx, user=default_value, number=default_value)`
            - one required and one optional parameter "user", "text":
                `async def command(ctx, user, text=default_value)`

            Note: Replace `default_value` with a value you want to be used if the parameter is not specified in discord, if you don't want a default value, just set it to `None`
            """
            base = format_name(base_names[0])
            sub = None
            if len(base_names) > 1:
                sub = format_name(base_names[1])
            if self.subcommands.get(base) is None:
                self.subcommands[base] = {}
            if sub is not None and self.subcommands[base].get(sub) is None:
                self.subcommands[base][sub] = {}
            
            command = SlashSubcommand(callback, base_names, name, description, options=options, guild_ids=guild_ids, default_permission=default_permission, guild_permissions=guild_permissions)
            self._add_to_cache(command)

        return wrapper
    def user_command(self, name=None, guild_ids=None, default_permission=True, guild_permissions=None):
        """
        Decorator for user context commands in discord.
            ``Right-click username`` -> ``apps`` -> ``commands is displayed here``


        Parameters
        ----------
            name: :class:`str`, optional
                The name of the command; default MISSING
            guild_ids: List[:class:`str` | :class:`int`]
                A list of guilds where the command can be used
            default_permission: :class:`bool`, optional
                Whether the command can be used by everyone; default True
            guild_permissions: Dict[:class:`SlashPermission`], optional
                Special permissions for guilds; default MISSING

        Decorator
        ---------

            callback: :class:`method(ctx, user)`
                The asynchron function that will be called if the command was used
                    ctx: :class:`~SlashedSubCommand`
                        The used slash command
                    user: :class:`discord.Member`
                        The user on which the command was used
                    
                    .. note::

                        ``ctx`` and ``user`` are just example names, you can use whatever you want for that

        Example
        -------
        
        .. code-block::

            @slash.user_command(name="call", guild_ids=[785567635802816595], default_permission=False, guild_permissions={
                785567635802816595: SlashPermission(allowed={
                    "585567635802816595": SlashPermission.USER
                })
            })
            async def call(ctx, message):
                ...
        """
        def wraper(callback):
            self._add_to_cache(UserCommand(callback, name, guild_ids, default_permission, guild_permissions))
        return wraper
    def message_command(self, name=None, guild_ids=None, default_permission=True, guild_permissions=None):
        """
        Decorator for message context commands in discord.
            ``Right-click message`` -> ``apps`` -> ``commands is displayed here``


        Parameters
        ----------
            name: :class:`str`, optional
                The name of the command; default MISSING
            guild_ids: List[:class:`str` | :class:`int`]
                A list of guilds where the command can be used
            default_permission: :class:`bool`, optional
                Whether the command can be used by everyone; default True
            guild_permissions: Dict[:class:`SlashPermission`], optional
                Special permissions for guilds; default MISSING

        Decorator
        ---------

            callback: :class:`method(ctx, message)`
                The asynchron function that will be called if the command was used
                    ctx: :class:`~SlashedSubCommand`
                        The used slash command
                    message: :class:`~Message`
                        The message on which the command was used
              
                    .. note::

                        ``ctx`` and ``message`` are just example names, you can use whatever you want for that
        
        Example
        -------
        
        .. code-block::

            @slash.message_command(name="quote", guild_ids=[785567635802816595], default_permission=False, guild_permissions={
                785567635802816595: SlashPermission(allowed={
                    "585567635802816595": SlashPermission.USER
                })
            })
            async def quote(ctx, message):
                ...
        """
        def wraper(callback):
            self._add_to_cache(MessageCommand(callback, name, guild_ids, default_permission, guild_permissions))
        return wraper

class Components():
    """
    A class for using and receiving message components in discord
    
    Parameters
    -----------
        client: :class:`discord.Client`
            The main discord client

        override_dpy: :class:`bool`
            Whether some of discord.py's default methods should be overriden with this libary's; Default ``True``
                For more information see https://github.com/discord-py-ui/discord-ui/blob/main/discord_ui/override.py

        auto_defer: Tuple[:class:`bool`, :class:`bool`]
            Settings for the auto-defer; Default ``(True, False)``

            ``[0]``: Whether interactions should be deferred automatically

            ``[1]``: Whether the deferration should be hidden (True) or public (False)

    Example
    ------------------
    Example for using the listener

    
    .. code-block::

        ...
        # Your bot declaration should be here
        components = Components(client)
        
    
    for listening to button presses, use
    
    .. code-block::

        ...
        @client.event("on_button_press")
        async def on_button(pressedButton):
            ...


    for listening to select menu selections, use

    .. code-block::

        ...
        @client.event("on_menu_select")
        async def on_select(seletedMenu):
            ...

    For components that will listen to a custom id, use

    .. code-block::

        ...
        @components.listening_component(custom_id="custom_id_here")
        async def my_func(ctx):
            ...

    """
    def __init__(self, client: com.Bot, override_dpy=True, auto_defer=False):
        """
        Creates a new compnent listener
        
        Example
        ```py
        Components(client, auto_defer=(True, False))
        ```
        """
        if override_dpy:
            override_it()

        # disable decompressing
        # self._buffer = bytearray()
        # self._zlib = zlib.decompressobj()

        self.auto_defer: Tuple[bool, bool] = (auto_defer, False) if isinstance(auto_defer, bool) else auto_defer
        self.listening_components: Dict[str, List[ListeningComponent]] = {}
        """A list of components that are listening for interaction"""
        self._discord: com.Bot = client
        self._discord._connection._component_listeners = {}
        if discord.__version__.startswith("2"):
            self._discord.add_listener(self._on_component_response, "on_socket_raw_receive")
        elif discord.__version__.startswith("1"):
            self._discord.add_listener(self._on_component_response, 'on_socket_response')

        old_add = self._discord.add_cog
        def add_cog_override(*args, **kwargs):
            cog = args[0] if len(args) > 0 else kwargs.get("cog")
            for com in self._get_listening_cogs(cog):
                com.cog = cog
                if self.listening_components.get(com.custom_id) is None:
                    self.listening_components[com.custom_id] = []
                self.listening_components[com.custom_id].append(com)
            old_add(*args, **kwargs)
        self._discord.add_cog = add_cog_override

        old_remove = self._discord.remove_cog
        def remove_cog_override(*args, **kwargs):
            cog = args[0] if len(args) > 0 else kwargs.get("cog")
            for com in self._get_listening_cogs(cog):
                com.cog = cog
                self.remove_listening_component(com)
            old_remove(*args, **kwargs)
        self._discord.remove_cog = remove_cog_override
    
    async def _on_component_response(self, msg):
        if discord.__version__.startswith("2"):
        # disable deecompressing
            if isinstance(msg, bytes):
                raise NotImplementedError("decompressing was removed! Please upgrade your discord.py version")
        #         self._buffer.extend(msg)

        #         if len(msg) < 4 or msg[-4:] != b'\x00\x00\xff\xff':
        #             return
        #         msg = self._zlib.decompress(self._buffer)
        #         msg = msg.decode('utf-8')
        #         self._buffer = bytearray()
            if isinstance(msg, str):
                msg = json.loads(msg)
        
        if msg["t"] != "INTERACTION_CREATE":
            return
        data = msg["d"]
        
        if data["type"] != 3:
            return
        
        user = discord.Member(data=data["member"], guild=self._discord._connection._get_guild(int(data["guild_id"])), state=self._discord._connection) if data.get("member") is not None else discord.User(state=self._discord._connection, data=data["user"])
        if user.dm_channel is None:
            await user.create_dm()
        msg = await getMessage(self._discord._connection, data=data, response=True)
        
        interaction = Interaction(self._discord._connection, data, user, msg)
        if self.auto_defer[0] is True:
            await interaction.defer(self.auto_defer[1])
        self._discord.dispatch("interaction_received", interaction)

        self._discord.dispatch("component", ComponentContext(self._discord._connection, data, user, msg))


        # Handle auto_defer
        if int(data["data"]["component_type"]) == 2:
            for x in msg.buttons:
                if hasattr(x, 'custom_id') and x.custom_id == data["data"]["custom_id"]:
                    component = PressedButton(data, user, x, msg, self._discord)
        elif int(data["data"]["component_type"]) == 3:
            for x in msg.select_menus:
                if x.custom_id == data["data"]["custom_id"]:
                    component = SelectedMenu(data, user, x, msg, self._discord)
        component._handle_auto_defer(self.auto_defer)
        
        # Get listening components with the same custom id
        listening_components = self.listening_components.get(data["data"]["custom_id"])
        if listening_components is not None:
            for listening_component in listening_components:
                await listening_component.invoke(component)

        listener: Listener = self._discord._connection._component_listeners.get(str(msg.id))
        if listener is not None:
            await listener._call_listeners(component)

        if ComponentType(data["data"]["component_type"]) is ComponentType.Button:
            self._discord.dispatch("button_press", component)
        elif ComponentType(data["data"]["component_type"]) is ComponentType.Select:
            self._discord.dispatch("menu_select", component)

    async def send(self, channel, content=MISSING, *, tts=False, embed=MISSING, embeds=MISSING, file=MISSING, 
            files=MISSING, delete_after=MISSING, nonce=MISSING, allowed_mentions=MISSING, reference=MISSING, 
            mention_author=MISSING, components=MISSING) -> Message:
        """
        Sends a message to a textchannel

        Parameters
        ----------
        channel: :class:`discord.TextChannel` | :class:`int` | :class:`str`
            The target textchannel or the id of it
        content: :class:`str`, optional
            The message text content; default None
        tts: :class:`bool`, optional
            True if this is a text-to-speech message; default False
        embed: :class:`discord.Message`, optional
            Embedded rich content (up to 6000 characters)
        embeds: List[:class:`discord.Embed`], optional
            Up to 10 embeds; default None
        file: :class:`discord.File`, optional
            A file sent as an attachment to the message; default None
        files: List[:class:`discord.File`], optional
            A list of file attachments; default None
        delete_after: :class:`float`, optional
            After how many seconds the message should be deleted; default None
        nonce: :class:`int`, optional
            The nonce to use for sending this message. If the message was successfully sent, then the message will have a nonce with this value; default None
        allowed_mentions: :class:`discord.AllowedMentions`, optional
            A list of mentions proceeded in the message; default None
        reference: :class:`discord.MessageReference` | :class:`discord.Message`, optional
            A message to refer to (reply); default None
        mention_author: :class:`bool`, optional
            True if the author should be mentioned; default None
        components: List[:class:`~Button` | :class:`~LinkButton` | :class:`~SelectMenu`], optional
            A list of message components included in this message; default None

        Returns
        -------
        :return: Returns the sent message
        :type: :class:`~Message`
        """

        if not isinstance(channel, (discord.TextChannel, int, str, discord.User)):
            raise WrongType("channel", channel, ["discord.TextChannel", "discord.User"])

        channel_id = None
        if isinstance(channel, discord.User):
            if channel.dm_channel is None:
                channel = await channel.create_dm()
                channel_id = channel.id
            else:
                channel_id = channel.dm_channel
        elif isinstance(channel, discord.TextChannel):
            channel_id = channel.id
        else: 
            channel_id = channel
        payload = jsonifyMessage(content=content, tts=tts, embed=embed, embeds=embeds, nonce=nonce, allowed_mentions=allowed_mentions, reference=reference, mention_author=mention_author, components=components)

        route = BetterRoute("POST", f"/channels/{channel_id}/messages")

        r = None
        if file is MISSING and files is MISSING:
            r = await self._discord.http.request(route, json=payload)
        else:
            r = await send_files(route, files=_or(files, [file]), payload=payload, http=self._discord.http)

        msg = Message(state=self._discord._connection, channel=channel, data=r)
        
        if not _none(delete_after):
            await msg.delete(delay=delete_after)
        
        return msg
    def send_webhook(self, webhook, content=MISSING, *, wait=False, username=MISSING, avatar_url=MISSING, tts=False, files=MISSING, embed=MISSING, embeds=MISSING, allowed_mentions=MISSING, components=MISSING) -> Union[WebhookMessage, None]:
        """
        Sends a webhook message
        
        Parameters
        ----------
            webhook: :class:`discord.Webhook`
                The webhook which will send the message
            content: :class:`str`, optional
                the message contents (up to 2000 characters); default None
            wait: :class:`bool`, optional
                if `True`, waits for server confirmation of message send before response, and returns the created message body; default False
            username: :class:`str`, optional
                override the default username of the webhook; default None
            avatar_url: :class:`str`, optional
                override the default avatar of the webhook; default None
            tts: :class:`bool`, optional
                true if this is a TTS message; default False
            files: :class:`discord.File`
                A list of files which will be sent as attachment
            embed: :class:`discord.Embed`
                Embed rich content, optional
            embeds: List[:class:`discord.Embed`], optional
                embedded rich content; default None
            allowed_mentions: :class:`discord.AllowedMentions`, optional
                allowed mentions for the message; default None
            components: List[:class:`~Button` | :class:`~LinkButton` | :class:`~SelectMenu`], optional
                the message components to include with the message; default None
        
        Returns
        -------
            :returns: The message sent, if wait was True, else nothing will be returned
            :type: :class:`~WebhookMessage` | :class:`None`
        
        """
        payload = jsonifyMessage(content, tts=tts, embed=embed, embeds=embeds, allowed_mentions=allowed_mentions, components=components)
        payload["wait"] = wait
        if username is not None:
            payload["username"] = username
        if avatar_url is not None:
            payload["avatar_url"] = str(avatar_url)

        return webhook._adapter.execute_webhook(payload=payload, wait=wait, files=files)
    def listening_component(self, custom_id, messages=None, users=None, component_type: Literal["button", "select"]=None, check=lambda component: True):
        """
        Decorator for ``add_listening_component``

        Parameters
        ----------
            custom_id: :class:`str`
                The custom_id of the components to listen to
            messages: List[:class:`discord.Message` | :class:`int` :class:`str`], Optional
                A list of messages or message ids to filter the listening component
            users: List[:class:`discord.User` | :class:`discord.Member` | :class:`int` | :class:`str`], Optional
                A list of users or user ids to filter
            omponent_type: Literal[``'button'`` | ``'select'``]
                What type the used component has to be of (select: SelectMenu, button: Button)
            check: :class:`function`, Optional
                A function that has to return True in order to invoke the listening component
                    The check function takes to parameters, the component and the message

        Decorator
        ---------
            callback: :class:`method(ctx)`
                The asynchron function that will be called if a component with the custom_id was invoked

                There will be one parameters passed

                    ctx: :class:`~PressedButton` or :class:`~SelectedMenu`
                        The invoked component
                    
                    .. note::

                        ``ctx`` is just an example name, you can use whatever you want for it

        Example
        -------
        .. code-block::

            @ui.components.listening_component("custom_id", [539459006847254542], [53945900682362362])
            async def callback(ctx):
                ...
            
        """
        def wrapper(callback):
            self.add_listening_component(callback, custom_id, messages, users, component_type, check)
        return wrapper
    def add_listening_component(self, callback, custom_id, messages=None, users=None, component_type: Literal["button", 2, "select", 3]=None, check=lambda component: True):
        """
        Adds a listener to received components

        Parameters
        ----------
            callback: :class:`function`
                The callback function that will be called when the component was received
            custom_id: :class:`str`
                The custom_id of the components to listen to
            messages: List[:class:`discord.Message` | :class:`int` :class:`str`], Optional
                A list of messages or message ids to filter the listening component
            users: List[:class:`discord.User` | :class:`discord.Member` | :class:`int` | :class:`str`], Optional
                A list of users or user ids to filter
            component_type: :class:`str` | :class:`int`
                The type of which the component has to be
            check: :class:`function`, Optional
                A function that has to return True in order to invoke the listening component
                    The check function takes to parameters, the component and the message

        """
        if not inspect.iscoroutinefunction(callback):
            raise NoAsyncCallback()
        if len(inspect.signature(callback).parameters) < 1:
            raise MissingListenedComponentParameters()
        
        if self.listening_components.get(custom_id) is None:
            self.listening_components[custom_id] = []
        self.listening_components[custom_id].append(ListeningComponent(callback, messages, users, component_type, check, custom_id))
    def remove_listening_components(self, custom_id):
        """
        Removes all listening components for a custom_id
        
        Parameters
        ----------
            custom_id: :class:`str`
                The custom_id for the listening component
        
        """
        if self.listening_components.get(custom_id) is not None:
            del self.listening_components[custom_id]
    def remove_listening_component(self, listening_component):
        """
        Removes a listening component
        
        Parameters
        ----------
            listening_component: :class:`ListeningComponent`
                The listening component which should be removed
        
        """
        with contextlib.supress(KeyError): 
            self.listening_components[listening_component.custom_id].remove(listening_component)

    def _get_listening_cogs(self, cog):
        return [x[1] for x in inspect.getmembers(cog, lambda x: isinstance(x, ListeningComponent))]

    async def put_listener_to(self, target_message, listener):
        """Adds a listener to a message and edits it if the components are missing
        
        Parameters
        ----------
        target_message: :class:`Message`
            The message to which the listener should be attached
        listener: :class:`Listener`
            The listener which should be put to the message
        
        """
        if len(target_message.components) == 0:
            await target_message.edit(components=listener.to_components())
        self.attach_listener_to(target_message, listener)

    def attach_listener_to(self, target_message, listener):
        """Attaches a listener to a message after it was sent
        
        Parameters
        ----------
        target_message: :class:`Message`
            The message to which the listener should be attached
        listener: :class:`Listener`
            The listener that will be attached
        
        """
        listener._start(target_message, target_message)
    def clear_listeners(self):
        """Removes all component listeners"""
        self._connection._component_listeners = {}

class UI():
    """
    The main extension for the package to use slash commands and message components
        
        Parameters
        ----------
            client: :class:`discord.ext.commands.Bot`
                The discord bot client

            override_dpy: :class:`bool`
                Whether some of discord.py's default methods should be overriden with this libary's; Default ``True``
                    For more information see https://github.com/discord-py-ui/discord-ui/blob/main/discord_ui/override.py

            slash_options: :class:`dict`, optional
                Settings for the slash command part; Default `{parse_method: ParseMethod.AUTO, delete_unused: False, wait_sync: 1}`
                
                ``parse_method``: :class:`int`, optional
                    How the received interaction argument data should be treated; Default ``ParseMethod.AUTO``

                ``auto_sync``: :class:`bool`, optional
                    Whether the libary should sync the slash commands automatically; Default ``True``

                ``delete_unused``: :class:`bool`, optional
                    Whether the commands that are not registered by this slash ui should be deleted in the api; Default ``False``

                ``sync_on_cog``: :class:`bool`, optional
                    Whether the slashcommands should be updated whenever a new cog is added or removed; Default ``False``

                ``wait_sync``: :class:`float`, optional
                    How many seconds will be waited until the commands are going to be synchronized; Default ``1``

        auto_defer: Tuple[:class:`bool`, :class:`bool`]
            Settings for the auto-defer; Default ``(True, False)``

            ``[0]``: Whether interactions should be deferred automatically

            ``[1]``: Whether the deferration should be hidden (True) or public (False)
    """
    def __init__(self, client, override_dpy=True, slash_options = {"parse_method": ParseMethod.AUTO, "auto_sync": True, "delete_unused": False, "sync_on_cog": True, "wait_sync": 1}, auto_defer = False) -> None:
        """
        Creates a new ui object
        
        Example
        ```py
        UI(client, slash_options={"delete_unused": True, "wait_sync": 2}, auto_defer=True)
        ```
        """
        self.components: Components = Components(client, override_dpy=override_dpy, auto_defer=auto_defer)
        """
        For using message components
        
        :type: :class:`~Components`
        """
        if slash_options is None:
            slash_options = {"resolve_data": True, "delete_unused": False, "wait_sync": 1, "auto_defer": auto_defer}
        if slash_options.get("auto_defer") is None:
            slash_options["auto_defer"] = auto_defer
        self.slash: Slash = Slash(client, **slash_options)
        """
        For using slash commands
        
        :type: :class:`~Slash`
        """
